# Copyright (c) Recommenders contributors.
# Licensed under the MIT License.


import pytest

from recommenders.utils.notebook_utils import execute_notebook, read_notebook


TOL = 0.05
ABS_TOL = 0.05


@pytest.mark.notebooks
def test_sar_single_node_smoke(notebooks, output_notebook, kernel_name):
    notebook_path = notebooks["sar_single_node"]
    execute_notebook(
        notebook_path,
        output_notebook,
        kernel_name=kernel_name,
        parameters=dict(TOP_K=10, MOVIELENS_DATA_SIZE="100k"),
    )
    results = read_notebook(output_notebook)

    assert results["map"] == pytest.approx(0.110591, rel=TOL, abs=ABS_TOL)
    assert results["ndcg"] == pytest.approx(0.382461, rel=TOL, abs=ABS_TOL)
    assert results["precision"] == pytest.approx(0.330753, rel=TOL, abs=ABS_TOL)
    assert results["recall"] == pytest.approx(0.176385, rel=TOL, abs=ABS_TOL)


@pytest.mark.notebooks
def test_baseline_deep_dive_smoke(notebooks, output_notebook, kernel_name):
    notebook_path = notebooks["baseline_deep_dive"]
    execute_notebook(
        notebook_path,
        output_notebook,
        kernel_name=kernel_name,
        parameters=dict(TOP_K=10, MOVIELENS_DATA_SIZE="100k"),
    )
    results = read_notebook(output_notebook)

    assert results["rmse"] == pytest.approx(1.054252, rel=TOL, abs=ABS_TOL)
    assert results["mae"] == pytest.approx(0.846033, rel=TOL, abs=ABS_TOL)
    assert results["rsquared"] == pytest.approx(0.136435, rel=TOL, abs=ABS_TOL)
    assert results["exp_var"] == pytest.approx(0.136446, rel=TOL, abs=ABS_TOL)
    assert results["map"] == pytest.approx(0.052850, rel=TOL, abs=ABS_TOL)
    assert results["ndcg"] == pytest.approx(0.248061, rel=TOL, abs=ABS_TOL)
    assert results["precision"] == pytest.approx(0.223754, rel=TOL, abs=ABS_TOL)
    assert results["recall"] == pytest.approx(0.108826, rel=TOL, abs=ABS_TOL)


@pytest.mark.notebooks
def test_surprise_svd_smoke(notebooks, output_notebook, kernel_name):
    notebook_path = notebooks["surprise_svd_deep_dive"]
    execute_notebook(
        notebook_path,
        output_notebook,
        kernel_name=kernel_name,
        parameters=dict(MOVIELENS_DATA_SIZE="100k"),
    )
    results = read_notebook(output_notebook)

    assert results["rmse"] == pytest.approx(0.96, rel=TOL, abs=ABS_TOL)
    assert results["mae"] == pytest.approx(0.75, rel=TOL, abs=ABS_TOL)
    assert results["rsquared"] == pytest.approx(0.29, rel=TOL, abs=ABS_TOL)
    assert results["exp_var"] == pytest.approx(0.29, rel=TOL, abs=ABS_TOL)
    assert results["map"] == pytest.approx(0.013, rel=TOL, abs=ABS_TOL)
    assert results["ndcg"] == pytest.approx(0.1, rel=TOL, abs=ABS_TOL)
    assert results["precision"] == pytest.approx(0.095, rel=TOL, abs=ABS_TOL)
    assert results["recall"] == pytest.approx(0.032, rel=TOL, abs=ABS_TOL)


@pytest.mark.notebooks
@pytest.mark.skip(reason="VW pip package has installation incompatibilities")
def test_vw_deep_dive_smoke(notebooks, output_notebook, kernel_name):
    notebook_path = notebooks["vowpal_wabbit_deep_dive"]
    execute_notebook(
        notebook_path,
        output_notebook,
        kernel_name=kernel_name,
        parameters=dict(MOVIELENS_DATA_SIZE="100k"),
    )
    results = read_notebook(output_notebook)

    assert results["rmse"] == pytest.approx(0.985920, rel=TOL, abs=ABS_TOL)
    assert results["mae"] == pytest.approx(0.71292, rel=TOL, abs=ABS_TOL)
    assert results["rsquared"] == pytest.approx(0.231199, rel=TOL, abs=ABS_TOL)
    assert results["exp_var"] == pytest.approx(0.231337, rel=TOL, abs=ABS_TOL)
    assert results["map"] == pytest.approx(0.012535, rel=TOL, abs=ABS_TOL)
    assert results["ndcg"] == pytest.approx(0.096594, rel=TOL, abs=ABS_TOL)
    assert results["precision"] == pytest.approx(0.097770, rel=TOL, abs=ABS_TOL)
    assert results["recall"] == pytest.approx(0.037612, rel=TOL, abs=ABS_TOL)


@pytest.mark.notebooks
def test_lightgbm_quickstart_smoke(notebooks, output_notebook, kernel_name):
    notebook_path = notebooks["lightgbm_quickstart"]
    execute_notebook(
        notebook_path,
        output_notebook,
        kernel_name=kernel_name,
        parameters=dict(
            MAX_LEAF=64,
            MIN_DATA=20,
            NUM_OF_TREES=100,
            TREE_LEARNING_RATE=0.15,
            EARLY_STOPPING_ROUNDS=20,
            METRIC="auc",
        ),
    )
    results = read_notebook(output_notebook)

    assert results["auc_basic"] == pytest.approx(0.7674, rel=TOL, abs=ABS_TOL)
    assert results["logloss_basic"] == pytest.approx(0.4669, rel=TOL, abs=ABS_TOL)
    assert results["auc_opt"] == pytest.approx(0.7757, rel=TOL, abs=ABS_TOL)
    assert results["logloss_opt"] == pytest.approx(0.4607, rel=TOL, abs=ABS_TOL)


@pytest.mark.notebooks
def test_cornac_bpr_smoke(notebooks, output_notebook, kernel_name):
    notebook_path = notebooks["cornac_bpr_deep_dive"]
    execute_notebook(
        notebook_path,
        output_notebook,
        kernel_name=kernel_name,
        parameters=dict(MOVIELENS_DATA_SIZE="100k"),
    )
    results = read_notebook(output_notebook)

    assert results["map"] == pytest.approx(0.1091, rel=TOL, abs=ABS_TOL)
    assert results["ndcg"] == pytest.approx(0.4034, rel=TOL, abs=ABS_TOL)
    assert results["precision"] == pytest.approx(0.3550, rel=TOL, abs=ABS_TOL)
    assert results["recall"] == pytest.approx(0.1802, rel=TOL, abs=ABS_TOL)



# Copyright (c) Recommenders contributors.
# Licensed under the MIT License.


import os
import pytest
from collections import defaultdict

try:
    from recommenders.models.sasrec.model import SASREC
    from recommenders.models.sasrec.ssept import SSEPT
    from recommenders.models.sasrec.sampler import WarpSampler
    from recommenders.models.sasrec.util import SASRecDataSet

    from recommenders.datasets.amazon_reviews import (
        download_and_extract,
        _reviews_preprocessing,
    )
except ImportError:
    pass  # skip if in cpu environment


@pytest.fixture()
def model_parameters():
    params = {
        "itemnum": 85930,
        "usernum": 63114,
        "maxlen": 50,
        "num_blocks": 2,
        "hidden_units": 100,
        "num_heads": 1,
        "dropout_rate": 0.1,
        "l2_emb": 0.0,
        "num_neg_test": 100,
    }
    return params


def data_process_with_time(fname, pname, K=10, sep=" ", item_set=None, add_time=False):
    User = defaultdict(list)
    Users = set()
    Items = set()
    user_dict, item_dict = {}, {}

    item_counter = defaultdict(lambda: 0)
    user_counter = defaultdict(lambda: 0)
    with open(fname, "r") as fr:
        for line in fr:
            u, i, t = line.rstrip().split(sep)
            User[u].append((i, t))
            Items.add(i)
            Users.add(u)
            item_counter[i] += 1
            user_counter[u] += 1

    # remove items with less than K interactions
    print(f"Read {len(User)} users and {len(Items)} items")
    remove_items = set()
    count_remove, count_missing = 0, 0
    for item in Items:
        if item_counter[item] < K:
            count_remove += 1
            remove_items.add(item)
        elif item_set and item not in item_set:
            count_missing += 1
            remove_items.add(item)

    if count_remove > 0:
        print(f"{count_remove} items have less than {K} interactions")

    if count_missing > 0:
        print(f"{count_missing} items are not in the meta data")

    Items = Items - remove_items

    # remove users with less than K interactions
    remove_users = set()
    count_remove = 0
    # Users = set(User.keys())
    for user in Users:
        if user_counter[user] < K:
            remove_users.add(user)
            count_remove += 1
    if count_remove > 0:
        print(f"{count_remove} users have less than {K} interactions")
        Users = Users - remove_users

    print(f"Total {len(Users)} users and {len(Items)} items")
    item_count = 1
    for item in Items:
        item_dict[item] = item_count
        item_count += 1

    count_del = 0
    user_count = 1
    with open(pname, "w") as fw:
        for user in Users:
            items = User[user]
            items = [tup for tup in items if tup[0] in Items]
            if len(items) < K:
                count_del += 1
            else:
                user_dict[user] = user_count
                # sort by time
                items = sorted(items, key=lambda x: x[1])

                # replace by the item-code
                timestamps = [x[1] for x in items]
                items = [item_dict[x[0]] for x in items]
                for i, t in zip(items, timestamps):
                    out_txt = [str(user_count), str(i)]
                    if add_time:
                        out_txt.append(str(t))
                    fw.write(sep.join(out_txt) + "\n")
                user_count += 1

    print(f"Total {user_count-1} users, {count_del} removed")
    print(f"Processed model input data in {pname}")
    return user_dict, item_dict


@pytest.mark.gpu
def test_prepare_data():
    data_dir = os.path.join("tests", "resources", "deeprec", "sasrec")
    dataset = "reviews_Electronics_5"
    reviews_name = dataset + ".json"
    outfile = os.path.join(data_dir, dataset + ".txt")

    reviews_file = os.path.join(data_dir, reviews_name)
    download_and_extract(reviews_name, reviews_file)
    reviews_output = _reviews_preprocessing(reviews_file)
    _, _ = data_process_with_time(reviews_output, outfile, K=10, sep="\t")

    # initiate a dataset class
    data = SASRecDataSet(filename=outfile, col_sep="\t")

    # create train, validation and test splits
    data.split()

    assert len(data.user_train) > 0
    assert len(data.user_valid) > 0
    assert len(data.user_test) > 0


@pytest.mark.gpu
def test_sampler():
    batch_size = 8
    maxlen = 50
    data_dir = os.path.join("tests", "resources", "deeprec", "sasrec")
    dataset = "reviews_Electronics_5"
    reviews_name = dataset + ".json"
    outfile = os.path.join(data_dir, dataset + ".txt")

    reviews_file = os.path.join(data_dir, reviews_name)
    download_and_extract(reviews_name, reviews_file)
    reviews_output = _reviews_preprocessing(reviews_file)
    _, _ = data_process_with_time(reviews_output, outfile, K=10, sep="\t")

    # initiate a dataset class
    data = SASRecDataSet(filename=outfile, col_sep="\t")

    # create train, validation and test splits
    data.split()

    sampler = WarpSampler(
        data.user_train,
        data.usernum,
        data.itemnum,
        batch_size=batch_size,
        maxlen=maxlen,
        n_workers=3,
    )
    u, seq, pos, neg = sampler.next_batch()

    assert len(u) == batch_size
    assert len(seq) == batch_size
    assert len(pos) == batch_size
    assert len(neg) == batch_size


@pytest.mark.gpu
def test_sasrec(model_parameters):
    params = model_parameters

    model = SASREC(
        item_num=params["itemnum"],
        seq_max_len=params["maxlen"],
        num_blocks=params["num_blocks"],
        embedding_dim=params["hidden_units"],
        attention_dim=params["hidden_units"],
        attention_num_heads=params["num_heads"],
        dropout_rate=params["dropout_rate"],
        conv_dims=[100, 100],
        l2_reg=params["l2_emb"],
        num_neg_test=params["num_neg_test"],
    )

    assert model.encoder is not None
    assert model.item_embedding_layer is not None


@pytest.mark.gpu
def test_ssept(model_parameters):
    params = model_parameters

    model = SSEPT(
        item_num=params["itemnum"],
        user_num=params["usernum"],
        seq_max_len=params["maxlen"],
        num_blocks=params["num_blocks"],
        user_embedding_dim=params["hidden_units"],
        item_embedding_dim=params["hidden_units"],
        attention_dim=params["hidden_units"],
        attention_num_heads=params["num_heads"],
        dropout_rate=params["dropout_rate"],
        conv_dims=[200, 200],
        l2_reg=params["l2_emb"],
        num_neg_test=params["num_neg_test"],
    )

    assert model.encoder is not None
    assert model.item_embedding_layer is not None
    assert model.user_embedding_layer is not None



# Copyright (c) Recommenders contributors.
# Licensed under the MIT License.

import random
import numpy as np
import pandas as pd
import scipy.sparse as sp

from recommenders.utils.constants import (
    DEFAULT_ITEM_COL,
    DEFAULT_USER_COL,
    DEFAULT_RATING_COL,
    DEFAULT_PREDICTION_COL,
)


class ImplicitCF(object):
    """Data processing class for GCN models which use implicit feedback.

    Initialize train and test set, create normalized adjacency matrix and sample data for training epochs.

    """

    def __init__(
        self,
        train,
        test=None,
        adj_dir=None,
        col_user=DEFAULT_USER_COL,
        col_item=DEFAULT_ITEM_COL,
        col_rating=DEFAULT_RATING_COL,
        col_prediction=DEFAULT_PREDICTION_COL,
        seed=None,
    ):
        """Constructor

        Args:
            adj_dir (str): Directory to save / load adjacency matrices. If it is None, adjacency
                matrices will be created and will not be saved.
            train (pandas.DataFrame): Training data with at least columns (col_user, col_item, col_rating).
            test (pandas.DataFrame): Test data with at least columns (col_user, col_item, col_rating).
                test can be None, if so, we only process the training data.
            col_user (str): User column name.
            col_item (str): Item column name.
            col_rating (str): Rating column name.
            seed (int): Seed.

        """
        self.user_idx = None
        self.item_idx = None
        self.adj_dir = adj_dir
        self.col_user = col_user
        self.col_item = col_item
        self.col_rating = col_rating
        self.col_prediction = col_prediction
        self.train, self.test = self._data_processing(train, test)
        self._init_train_data()

        random.seed(seed)

    def _data_processing(self, train, test):
        """Process the dataset to reindex userID and itemID and only keep records with ratings greater than 0.

        Args:
            train (pandas.DataFrame): Training data with at least columns (col_user, col_item, col_rating).
            test (pandas.DataFrame): Test data with at least columns (col_user, col_item, col_rating).
                test can be None, if so, we only process the training data.

        Returns:
            list: train and test pandas.DataFrame Dataset, which have been reindexed and filtered.

        """
        df = (
            train
            if test is None
            else pd.concat([train, test], axis=0, ignore_index=True)
        )

        if self.user_idx is None:
            user_idx = df[[self.col_user]].drop_duplicates().reindex()
            user_idx[self.col_user + "_idx"] = np.arange(len(user_idx))
            self.n_users = len(user_idx)
            self.n_users_in_train = train[self.col_user].nunique()
            self.user_idx = user_idx

            self.user2id = dict(
                zip(user_idx[self.col_user], user_idx[self.col_user + "_idx"])
            )
            self.id2user = dict(
                zip(user_idx[self.col_user + "_idx"], user_idx[self.col_user])
            )

        if self.item_idx is None:
            item_idx = df[[self.col_item]].drop_duplicates()
            item_idx[self.col_item + "_idx"] = np.arange(len(item_idx))
            self.n_items = len(item_idx)
            self.item_idx = item_idx

            self.item2id = dict(
                zip(item_idx[self.col_item], item_idx[self.col_item + "_idx"])
            )
            self.id2item = dict(
                zip(item_idx[self.col_item + "_idx"], item_idx[self.col_item])
            )

        return self._reindex(train), self._reindex(test)

    def _reindex(self, df):
        """Process the dataset to reindex userID and itemID and only keep records with ratings greater than 0.

        Args:
            df (pandas.DataFrame): dataframe with at least columns (col_user, col_item, col_rating).

        Returns:
            list: train and test pandas.DataFrame Dataset, which have been reindexed and filtered.

        """

        if df is None:
            return None

        df = pd.merge(df, self.user_idx, on=self.col_user, how="left")
        df = pd.merge(df, self.item_idx, on=self.col_item, how="left")

        df = df[df[self.col_rating] > 0]

        df_reindex = df[
            [self.col_user + "_idx", self.col_item + "_idx", self.col_rating]
        ]
        df_reindex.columns = [self.col_user, self.col_item, self.col_rating]

        return df_reindex

    def _init_train_data(self):
        """Record items interated with each user in a dataframe self.interact_status, and create adjacency
        matrix self.R.

        """
        self.interact_status = (
            self.train.groupby(self.col_user)[self.col_item]
            .apply(set)
            .reset_index()
            .rename(columns={self.col_item: self.col_item + "_interacted"})
        )
        self.R = sp.dok_matrix((self.n_users, self.n_items), dtype=np.float32)
        self.R[self.train[self.col_user], self.train[self.col_item]] = 1.0

    def get_norm_adj_mat(self):
        """Load normalized adjacency matrix if it exists, otherwise create (and save) it.

        Returns:
            scipy.sparse.csr_matrix: Normalized adjacency matrix.

        """
        try:
            if self.adj_dir is None:
                raise FileNotFoundError
            norm_adj_mat = sp.load_npz(self.adj_dir + "/norm_adj_mat.npz")
            print("Already load norm adj matrix.")

        except FileNotFoundError:
            norm_adj_mat = self.create_norm_adj_mat()
            if self.adj_dir is not None:
                sp.save_npz(self.adj_dir + "/norm_adj_mat.npz", norm_adj_mat)
        return norm_adj_mat

    def create_norm_adj_mat(self):
        """Create normalized adjacency matrix.

        Returns:
            scipy.sparse.csr_matrix: Normalized adjacency matrix.

        """
        adj_mat = sp.dok_matrix(
            (self.n_users + self.n_items, self.n_users + self.n_items), dtype=np.float32
        )
        adj_mat = adj_mat.tolil()
        R = self.R.tolil()

        adj_mat[: self.n_users, self.n_users :] = R
        adj_mat[self.n_users :, : self.n_users] = R.T
        adj_mat = adj_mat.todok()
        print("Already create adjacency matrix.")

        rowsum = np.array(adj_mat.sum(1))
        d_inv = np.power(rowsum + 1e-9, -0.5).flatten()
        d_inv[np.isinf(d_inv)] = 0.0
        d_mat_inv = sp.diags(d_inv)
        norm_adj_mat = d_mat_inv.dot(adj_mat)
        norm_adj_mat = norm_adj_mat.dot(d_mat_inv)
        print("Already normalize adjacency matrix.")

        return norm_adj_mat.tocsr()

    def train_loader(self, batch_size):
        """Sample train data every batch. One positive item and one negative item sampled for each user.

        Args:
            batch_size (int): Batch size of users.

        Returns:
            numpy.ndarray, numpy.ndarray, numpy.ndarray:
            - Sampled users.
            - Sampled positive items.
            - Sampled negative items.
        """

        def sample_neg(x):
            if len(x) >= self.n_items:
                raise ValueError("A user has voted in every item. Can't find a negative sample.")
            while True:
                neg_id = random.randint(0, self.n_items - 1)
                if neg_id not in x:
                    return neg_id

        indices = range(self.n_users_in_train)
        if self.n_users < batch_size:
            users = [random.choice(indices) for _ in range(batch_size)]
        else:
            users = random.sample(indices, batch_size)

        interact = self.interact_status.iloc[users]
        pos_items = interact[self.col_item + "_interacted"].apply(
            lambda x: random.choice(list(x))
        )
        neg_items = interact[self.col_item + "_interacted"].apply(
            lambda x: sample_neg(x)
        )

        return np.array(users), np.array(pos_items), np.array(neg_items)



# Copyright (c) Recommenders contributors.
# Licensed under the MIT License.


import numpy as np
import pandas as pd
import pytest
from unittest.mock import Mock
from sklearn.preprocessing import minmax_scale
from pandas.testing import assert_frame_equal

from recommenders.utils.constants import (
    DEFAULT_USER_COL,
    DEFAULT_ITEM_COL,
    DEFAULT_RATING_COL,
    DEFAULT_PREDICTION_COL,
)
from recommenders.evaluation.python_evaluation import (
    _check_column_dtypes,
    merge_rating_true_pred,
    merge_ranking_true_pred,
    rmse,
    mae,
    rsquared,
    exp_var,
    get_top_k_items,
    precision_at_k,
    r_precision_at_k,
    recall_at_k,
    ndcg_at_k,
    map_at_k,
    map,
    auc,
    logloss,
    user_diversity,
    diversity,
    historical_item_novelty,
    novelty,
    user_item_serendipity,
    user_serendipity,
    serendipity,
    catalog_coverage,
    distributional_coverage,
    ColumnMismatchError,
    ColumnTypeMismatchError,
)


TOL = 0.0001


@pytest.fixture
def rating_true_binary(rating_true):
    # Convert true ratings to binary
    rating_true[DEFAULT_RATING_COL] = rating_true[DEFAULT_RATING_COL].apply(
        lambda x: 1.0 if x >= 3 else 0.0
    )
    return rating_true


@pytest.fixture
def rating_pred_binary(rating_pred):
    # Normalize the predictions
    rating_pred[DEFAULT_PREDICTION_COL] = minmax_scale(
        rating_pred[DEFAULT_PREDICTION_COL].astype(float)
    )
    return rating_pred


def test_column_dtypes_match(rating_true, rating_pred):
    # Check if _check_column_dtypes wraps the input function (i.e. returns the same output as the input function's)
    mocked_fn = Mock(return_value=True)
    assert _check_column_dtypes(mocked_fn)(
        rating_true,
        rating_pred,
        col_user=DEFAULT_USER_COL,
        col_item=DEFAULT_ITEM_COL,
        col_rating=DEFAULT_RATING_COL,
        col_prediction=DEFAULT_PREDICTION_COL,
    )

    # Change data types of true and prediction data, and there should type mismatch error produced
    rating_true[DEFAULT_USER_COL] = rating_true[DEFAULT_USER_COL].astype(str)
    with pytest.raises(ColumnTypeMismatchError):
        _check_column_dtypes(Mock())(
            rating_true,
            rating_pred,
            col_user=DEFAULT_USER_COL,
            col_item=DEFAULT_ITEM_COL,
            col_rating=DEFAULT_RATING_COL,
            col_prediction=DEFAULT_PREDICTION_COL,
        )

    # Drop a column, and there should column mismatch error produced
    rating_true.drop(DEFAULT_USER_COL, axis="columns", inplace=True)
    with pytest.raises(ColumnMismatchError):
        _check_column_dtypes(Mock())(
            rating_true,
            rating_pred,
            col_user=DEFAULT_USER_COL,
            col_item=DEFAULT_ITEM_COL,
            col_rating=DEFAULT_RATING_COL,
            col_prediction=DEFAULT_PREDICTION_COL,
        )


def test_merge_rating(rating_true, rating_pred):
    y_true, y_pred = merge_rating_true_pred(
        rating_true,
        rating_pred,
        col_user=DEFAULT_USER_COL,
        col_item=DEFAULT_ITEM_COL,
        col_rating=DEFAULT_RATING_COL,
        col_prediction=DEFAULT_PREDICTION_COL,
    )
    target_y_true = np.array([3, 3, 5, 5, 3, 3, 2, 1])
    target_y_pred = np.array([14, 12, 7, 8, 13, 6, 11, 5])

    assert y_true.shape == y_pred.shape
    assert np.all(y_true == target_y_true)
    assert np.all(y_pred == target_y_pred)


def test_merge_ranking(rating_true, rating_pred):

    data_hit, data_hit_count, n_users = merge_ranking_true_pred(
        rating_true,
        rating_pred,
        col_user=DEFAULT_USER_COL,
        col_item=DEFAULT_ITEM_COL,
        col_rating=DEFAULT_RATING_COL,
        col_prediction=DEFAULT_PREDICTION_COL,
        relevancy_method="top_k",
    )

    assert isinstance(data_hit, pd.DataFrame)

    assert isinstance(data_hit_count, pd.DataFrame)
    columns = data_hit_count.columns
    columns_exp = [DEFAULT_USER_COL, DEFAULT_ITEM_COL, DEFAULT_PREDICTION_COL]
    assert set(columns).intersection(set(columns_exp)) is not None

    assert n_users == 3


def test_python_rmse(rating_true, rating_pred):
    assert (
        rmse(
            rating_true=rating_true,
            rating_pred=rating_true,
            col_prediction=DEFAULT_RATING_COL,
        )
        == 0
    )
    assert rmse(rating_true, rating_pred) == pytest.approx(7.254309, TOL)


def test_python_mae(rating_true, rating_pred):
    assert (
        mae(
            rating_true=rating_true,
            rating_pred=rating_true,
            col_prediction=DEFAULT_RATING_COL,
        )
        == 0
    )
    assert mae(rating_true, rating_pred) == pytest.approx(6.375, TOL)


def test_python_rsquared(rating_true, rating_pred):
    assert rsquared(
        rating_true=rating_true,
        rating_pred=rating_true,
        col_prediction=DEFAULT_RATING_COL,
    ) == pytest.approx(1.0, TOL)
    assert rsquared(rating_true, rating_pred) == pytest.approx(-31.699029, TOL)


def test_python_exp_var(rating_true, rating_pred):
    assert exp_var(
        rating_true=rating_true,
        rating_pred=rating_true,
        col_prediction=DEFAULT_RATING_COL,
    ) == pytest.approx(1.0, TOL)
    assert exp_var(rating_true, rating_pred) == pytest.approx(-6.4466, TOL)


def test_get_top_k_items(rating_true):
    top_3_items_df = get_top_k_items(
        dataframe=rating_true,
        col_user=DEFAULT_USER_COL,
        col_rating=DEFAULT_RATING_COL,
        k=3,
    )
    top_3_user_true = pd.Series([1, 1, 1, 2, 2, 2, 3, 3, 3])
    top_3_rating_true = pd.Series([5, 4, 3, 5, 5, 3, 5, 5, 5])
    top_3_rank_true = pd.Series([1, 2, 3, 1, 2, 3, 1, 2, 3])
    assert top_3_items_df[DEFAULT_USER_COL].equals(top_3_user_true)
    assert top_3_items_df[DEFAULT_RATING_COL].equals(top_3_rating_true)
    assert top_3_items_df["rank"].equals(top_3_rank_true)
    assert top_3_items_df[DEFAULT_ITEM_COL][:3].equals(pd.Series([1, 2, 3]))
    # First two itemIDs of user 2. The scores are both 5, so any order is OK.
    assert set(top_3_items_df[DEFAULT_ITEM_COL][3:5]) == set([1, 4])
    # Third itemID of user 2. Both item 5 and 6 have a score of 3, so either one is OK.
    assert top_3_items_df[DEFAULT_ITEM_COL][5] in [5, 6]
    # All itemIDs of user 3. All three items have a score of 5, so any order is OK.
    assert set(top_3_items_df[DEFAULT_ITEM_COL][6:]) == set([2, 5, 6])


# Test get_top_k_items() when k is larger than the number of available items
def test_get_top_k_items_largek(rating_true):
    top_6_items_df = get_top_k_items(
        dataframe=rating_true,
        col_user=DEFAULT_USER_COL,
        col_rating=DEFAULT_RATING_COL,
        k=6,
    )
    top_6_user_true = pd.Series([1, 1, 1, 2, 2, 2, 2, 2, 3, 3, 3, 3, 3, 3])
    top_6_rating_true = pd.Series([5, 4, 3, 5, 5, 3, 3, 1, 5, 5, 5, 4, 4, 3])
    top_6_rank_true = pd.Series([1, 2, 3, 1, 2, 3, 4, 5, 1, 2, 3, 4, 5, 6])
    assert top_6_items_df[DEFAULT_USER_COL].equals(top_6_user_true)
    assert top_6_items_df[DEFAULT_RATING_COL].equals(top_6_rating_true)
    assert top_6_items_df["rank"].equals(top_6_rank_true)
    assert top_6_items_df[DEFAULT_ITEM_COL][:3].equals(pd.Series([1, 2, 3]))
    # First two itemIDs of user 2. The scores are both 5, so any order is OK.
    assert set(top_6_items_df[DEFAULT_ITEM_COL][3:5]) == set([1, 4])
    # Third and fourth itemID of user 2. The scores are both 3, so any order is OK.
    assert set(top_6_items_df[DEFAULT_ITEM_COL][5:7]) == set([5, 6])
    assert top_6_items_df[DEFAULT_ITEM_COL][7] == 7
    # First three itemIDs of user 3. The scores are both 5, so any order is OK.
    assert set(top_6_items_df[DEFAULT_ITEM_COL][8:11]) == set([2, 5, 6])
    # Fourth and fifth itemID of user 3. The scores are both 4, so any order is OK.
    assert set(top_6_items_df[DEFAULT_ITEM_COL][11:13]) == set([8, 9])
    # Sixth itemID of user 3. Item 10,11,12 have a score of 3, so either one is OK.
    assert top_6_items_df[DEFAULT_ITEM_COL][13] in [10, 11, 12]


def test_python_ndcg_at_k(rating_true, rating_pred, rating_nohit):
    assert ndcg_at_k(
        rating_true=rating_true,
        rating_pred=rating_true,
        col_prediction=DEFAULT_RATING_COL,
        k=10,
    ) == pytest.approx(1.0, TOL)
    assert ndcg_at_k(rating_true, rating_nohit, k=10) == 0.0
    assert ndcg_at_k(rating_true, rating_pred, k=10) == pytest.approx(0.38172, TOL)

    # Test raw relevance score and log2 discounting factor using wiki example
    # See https://en.wikipedia.org/wiki/Discounted_cumulative_gain
    df_true = pd.DataFrame(
        {
            DEFAULT_USER_COL: np.full(8, 0, dtype=int),
            DEFAULT_ITEM_COL: np.arange(8),
            DEFAULT_RATING_COL: np.asarray([3, 2, 3, 0, 1, 2, 3, 2]),
        }
    )
    df_pred = pd.DataFrame(
        {
            DEFAULT_USER_COL: np.full(6, 0, dtype=int),
            DEFAULT_ITEM_COL: np.arange(6),
            DEFAULT_PREDICTION_COL: np.asarray([6, 5, 4, 3, 2, 1]),
        }
    )
    assert ndcg_at_k(
        df_true, df_pred, k=6, score_type="raw", discfun_type="log2"
    ) == pytest.approx(0.785, TOL)


def test_python_map(rating_true, rating_pred, rating_nohit):
    assert (
        map(
            rating_true=rating_true,
            rating_pred=rating_true,
            col_prediction=DEFAULT_RATING_COL,
            k=10,
        )
        == 1
    )
    assert map(rating_true, rating_nohit, k=10) == 0.0
    assert map(rating_true, rating_pred, k=10) == pytest.approx(0.23613, TOL)


def test_python_map_at_k(rating_true, rating_pred, rating_nohit):
    assert (
        map_at_k(
            rating_true=rating_true,
            rating_pred=rating_true,
            col_prediction=DEFAULT_RATING_COL,
            k=10,
        )
        == 1
    )
    assert map_at_k(rating_true, rating_nohit, k=10) == 0.0
    assert map_at_k(rating_true, rating_pred, k=10) == pytest.approx(0.23613, TOL)


def test_python_precision_at_k(rating_true, rating_pred, rating_nohit):
    assert (
        precision_at_k(
            rating_true=rating_true,
            rating_pred=rating_true,
            col_prediction=DEFAULT_RATING_COL,
            k=10,
        )
        == 0.6
    )
    assert precision_at_k(rating_true, rating_nohit, k=10) == 0.0
    assert precision_at_k(rating_true, rating_pred, k=10) == pytest.approx(0.26666, TOL)

    # Check normalization
    single_user = pd.DataFrame(
        {
            DEFAULT_USER_COL: [1, 1, 1],
            DEFAULT_ITEM_COL: [1, 2, 3],
            DEFAULT_RATING_COL: [5, 4, 3],
        }
    )
    assert (
        precision_at_k(
            rating_true=single_user,
            rating_pred=single_user,
            col_rating=DEFAULT_RATING_COL,
            col_prediction=DEFAULT_RATING_COL,
            k=3,
        )
        == 1
    )

    same_items = pd.DataFrame(
        {
            DEFAULT_USER_COL: [1, 1, 1, 2, 2, 2],
            DEFAULT_ITEM_COL: [1, 2, 3, 1, 2, 3],
            DEFAULT_RATING_COL: [5, 4, 3, 5, 5, 3],
        }
    )
    assert (
        precision_at_k(
            rating_true=same_items,
            rating_pred=same_items,
            col_prediction=DEFAULT_RATING_COL,
            k=3,
        )
        == 1
    )

    # Check that if the sample size is smaller than k, the maximum precision can not be 1
    # if we do precision@5 when there is only 3 items, we can get a maximum of 3/5.
    assert (
        precision_at_k(
            rating_true=same_items,
            rating_pred=same_items,
            col_prediction=DEFAULT_RATING_COL,
            k=5,
        )
        == 0.6
    )


def test_python_recall_at_k(rating_true, rating_pred, rating_nohit):
    assert recall_at_k(
        rating_true=rating_true,
        rating_pred=rating_true,
        col_prediction=DEFAULT_RATING_COL,
        k=10,
    ) == pytest.approx(1, TOL)
    assert recall_at_k(rating_true, rating_nohit, k=10) == 0.0
    assert recall_at_k(rating_true, rating_pred, k=10) == pytest.approx(0.37777, TOL)


def test_python_r_precision(rating_true, rating_pred, rating_nohit):
    assert r_precision_at_k(
        rating_true=rating_true,
        rating_pred=rating_true,
        col_prediction=DEFAULT_RATING_COL,
        k=10,
    ) == pytest.approx(1, TOL)
    assert r_precision_at_k(rating_true, rating_nohit, k=5) == 0.0
    assert r_precision_at_k(rating_true, rating_pred, k=3) == pytest.approx(
        0.21111, TOL
    )
    assert r_precision_at_k(rating_true, rating_pred, k=5) == pytest.approx(
        0.24444, TOL
    )
    # Equivalent to precision
    assert r_precision_at_k(rating_true, rating_pred, k=10) == pytest.approx(
        0.37777, TOL
    )


def test_python_auc(rating_true_binary, rating_pred_binary):
    assert auc(
        rating_true=rating_true_binary,
        rating_pred=rating_true_binary,
        col_prediction=DEFAULT_RATING_COL,
    ) == pytest.approx(1.0, TOL)

    assert auc(
        rating_true=rating_true_binary,
        rating_pred=rating_pred_binary,
        col_rating=DEFAULT_RATING_COL,
        col_prediction=DEFAULT_PREDICTION_COL,
    ) == pytest.approx(0.75, TOL)


def test_python_logloss(rating_true_binary, rating_pred_binary):
    assert logloss(
        rating_true=rating_true_binary,
        rating_pred=rating_true_binary,
        col_prediction=DEFAULT_RATING_COL,
    ) == pytest.approx(0, TOL)

    assert logloss(
        rating_true=rating_true_binary,
        rating_pred=rating_pred_binary,
        col_rating=DEFAULT_RATING_COL,
        col_prediction=DEFAULT_PREDICTION_COL,
    ) == pytest.approx(0.7835, TOL)


def test_python_errors(rating_true, rating_pred):
    with pytest.raises(ColumnMismatchError):
        rmse(rating_true, rating_true, col_user="not_user")

    with pytest.raises(ColumnMismatchError):
        mae(
            rating_pred,
            rating_pred,
            col_rating=DEFAULT_PREDICTION_COL,
            col_user="not_user",
        )

    with pytest.raises(ColumnMismatchError):
        rsquared(rating_true, rating_pred, col_item="not_item")

    with pytest.raises(ColumnMismatchError):
        exp_var(
            rating_pred,
            rating_pred,
            col_rating=DEFAULT_PREDICTION_COL,
            col_item="not_item",
        )

    with pytest.raises(ColumnMismatchError):
        precision_at_k(rating_true, rating_pred, col_prediction="not_prediction")

    with pytest.raises(ColumnMismatchError):
        recall_at_k(rating_true, rating_pred, col_prediction="not_prediction")

    with pytest.raises(ColumnMismatchError):
        ndcg_at_k(rating_true, rating_true, col_user="not_user")

    with pytest.raises(ColumnMismatchError):
        map_at_k(
            rating_pred,
            rating_pred,
            col_rating=DEFAULT_PREDICTION_COL,
            col_user="not_user",
        )


def test_catalog_coverage(diversity_data):
    train_df, reco_df, _ = diversity_data
    c_coverage = catalog_coverage(
        train_df=train_df, reco_df=reco_df, col_user="UserId", col_item="ItemId"
    )
    assert c_coverage == pytest.approx(0.8, TOL)


def test_distributional_coverage(diversity_data):
    train_df, reco_df, _ = diversity_data
    d_coverage = distributional_coverage(
        train_df=train_df, reco_df=reco_df, col_user="UserId", col_item="ItemId"
    )
    assert d_coverage == pytest.approx(1.9183, TOL)


def test_item_novelty(diversity_data):
    train_df, reco_df, _ = diversity_data
    actual = historical_item_novelty(
        train_df=train_df, reco_df=reco_df, col_user="UserId", col_item="ItemId"
    )
    assert_frame_equal(
        pd.DataFrame(
            dict(ItemId=[1, 2, 3, 4, 5], item_novelty=[3.0, 3.0, 2.0, 1.41504, 3.0])
        ),
        actual,
        check_exact=False,
    )
    assert np.all(actual["item_novelty"].values >= 0)
    # Test that novelty is zero when data includes only one item
    train_df_new = train_df.loc[train_df["ItemId"] == 3]
    actual = historical_item_novelty(
        train_df=train_df_new, reco_df=reco_df, col_user="UserId", col_item="ItemId"
    )
    assert actual["item_novelty"].values[0] == 0


def test_novelty(diversity_data):
    train_df, reco_df, _ = diversity_data
    assert novelty(
        train_df=train_df, reco_df=reco_df, col_user="UserId", col_item="ItemId"
    ) == pytest.approx(2.83333, TOL)

    # Test that novelty is zero when data includes only one item
    train_df_new = train_df.loc[train_df["ItemId"] == 3]
    reco_df_new = reco_df.loc[reco_df["ItemId"] == 3]
    assert (
        novelty(
            train_df=train_df_new,
            reco_df=reco_df_new,
            col_user="UserId",
            col_item="ItemId",
        )
        == 0
    )


def test_user_diversity(diversity_data):
    train_df, reco_df, _ = diversity_data
    actual = user_diversity(
        train_df=train_df,
        reco_df=reco_df,
        item_feature_df=None,
        item_sim_measure="item_cooccurrence_count",
        col_user="UserId",
        col_item="ItemId",
        col_sim="sim",
        col_relevance=None,
    )
    assert_frame_equal(
        pd.DataFrame(dict(UserId=[1, 2, 3], user_diversity=[0.29289, 1.0, 0.0])),
        actual,
        check_exact=False,
        atol=TOL,
    )


def test_diversity(diversity_data):
    train_df, reco_df, _ = diversity_data
    assert diversity(
        train_df=train_df,
        reco_df=reco_df,
        item_feature_df=None,
        item_sim_measure="item_cooccurrence_count",
        col_user="UserId",
        col_item="ItemId",
        col_sim="sim",
        col_relevance=None,
    ) == pytest.approx(0.43096, TOL)


def test_user_item_serendipity(diversity_data):
    train_df, reco_df, _ = diversity_data
    actual = user_item_serendipity(
        train_df=train_df,
        reco_df=reco_df,
        item_feature_df=None,
        item_sim_measure="item_cooccurrence_count",
        col_user="UserId",
        col_item="ItemId",
        col_sim="sim",
        col_relevance="Relevance",
    )
    assert_frame_equal(
        pd.DataFrame(
            dict(
                UserId=[1, 1, 2, 2, 3, 3],
                ItemId=[3, 5, 2, 5, 1, 2],
                user_item_serendipity=[
                    0.72783,
                    0.0,
                    0.71132,
                    0.35777,
                    0.80755,
                    0.0,
                ],
            )
        ),
        actual,
        check_exact=False,
    )


def test_user_serendipity(diversity_data):
    train_df, reco_df, _ = diversity_data
    actual = user_serendipity(
        train_df=train_df,
        reco_df=reco_df,
        item_feature_df=None,
        item_sim_measure="item_cooccurrence_count",
        col_user="UserId",
        col_item="ItemId",
        col_sim="sim",
        col_relevance="Relevance",
    )
    assert_frame_equal(
        pd.DataFrame(
            dict(UserId=[1, 2, 3], user_serendipity=[0.363915, 0.53455, 0.403775])
        ),
        actual,
        check_exact=False,
    )


def test_serendipity(diversity_data):
    train_df, reco_df, _ = diversity_data
    assert serendipity(
        train_df=train_df,
        reco_df=reco_df,
        item_feature_df=None,
        item_sim_measure="item_cooccurrence_count",
        col_user="UserId",
        col_item="ItemId",
        col_sim="sim",
        col_relevance="Relevance",
    ) == pytest.approx(0.43408, TOL)


def test_user_diversity_item_feature_vector(diversity_data):
    train_df, reco_df, item_feature_df = diversity_data
    actual = user_diversity(
        train_df=train_df,
        reco_df=reco_df,
        item_feature_df=item_feature_df,
        item_sim_measure="item_feature_vector",
        col_user="UserId",
        col_item="ItemId",
        col_sim="sim",
        col_relevance=None,
    )
    assert_frame_equal(
        pd.DataFrame(dict(UserId=[1, 2, 3], user_diversity=[0.5000, 0.5000, 0.5000])),
        actual,
        check_exact=False,
    )


def test_diversity_item_feature_vector(diversity_data):
    train_df, reco_df, item_feature_df = diversity_data
    assert diversity(
        train_df=train_df,
        reco_df=reco_df,
        item_feature_df=item_feature_df,
        item_sim_measure="item_feature_vector",
        col_user="UserId",
        col_item="ItemId",
        col_sim="sim",
        col_relevance=None,
    ) == pytest.approx(0.5000, TOL)


def test_user_item_serendipity_item_feature_vector(
    diversity_data,
):
    train_df, reco_df, item_feature_df = diversity_data
    actual = user_item_serendipity(
        train_df=train_df,
        reco_df=reco_df,
        item_feature_df=item_feature_df,
        item_sim_measure="item_feature_vector",
        col_user="UserId",
        col_item="ItemId",
        col_sim="sim",
        col_relevance="Relevance",
    )
    assert_frame_equal(
        pd.DataFrame(
            dict(
                UserId=[1, 1, 2, 2, 3, 3],
                ItemId=[3, 5, 2, 5, 1, 2],
                user_item_serendipity=[
                    0.5000,
                    0.0,
                    0.75,
                    0.5000,
                    0.6667,
                    0.0,
                ],
            )
        ),
        actual,
        check_exact=False,
        atol=TOL,
    )


def test_user_serendipity_item_feature_vector(diversity_data):
    train_df, reco_df, item_feature_df = diversity_data
    actual = user_serendipity(
        train_df=train_df,
        reco_df=reco_df,
        item_feature_df=item_feature_df,
        item_sim_measure="item_feature_vector",
        col_user="UserId",
        col_item="ItemId",
        col_sim="sim",
        col_relevance="Relevance",
    )
    assert_frame_equal(
        pd.DataFrame(dict(UserId=[1, 2, 3], user_serendipity=[0.2500, 0.625, 0.3333])),
        actual,
        check_exact=False,
        atol=TOL,
    )


def test_serendipity_item_feature_vector(diversity_data):
    train_df, reco_df, item_feature_df = diversity_data
    assert serendipity(
        train_df=train_df,
        reco_df=reco_df,
        item_feature_df=item_feature_df,
        item_sim_measure="item_feature_vector",
        col_user="UserId",
        col_item="ItemId",
        col_sim="sim",
        col_relevance="Relevance",
    ) == pytest.approx(0.4028, TOL)



# Copyright (c) Recommenders contributors.
# Licensed under the MIT License.

import sys
import os
import glob
import logging
from numba import cuda
from numba.cuda.cudadrv.error import CudaSupportError


logger = logging.getLogger(__name__)


DEFAULT_CUDA_PATH_LINUX = "/usr/local/cuda/version.txt"


def get_number_gpus():
    """Get the number of GPUs in the system.
    Returns:
        int: Number of GPUs.
    """
    try:
        import torch

        return torch.cuda.device_count()
    except (ImportError, ModuleNotFoundError):
        pass
    try:
        import numba

        return len(numba.cuda.gpus)
    except Exception:  # numba.cuda.cudadrv.error.CudaSupportError:
        return 0


def get_gpu_info():
    """Get information of GPUs.

    Returns:
        list: List of gpu information dictionary as with `device_name`, `total_memory` (in Mb) and `free_memory` (in Mb).
        Returns an empty list if there is no cuda device available.
    """
    gpus = []
    try:
        for gpu in cuda.gpus:
            with gpu:
                meminfo = cuda.current_context().get_memory_info()
                g = {
                    "device_name": gpu.name.decode("ASCII"),
                    "total_memory": meminfo[1] / 1048576,  # Mb
                    "free_memory": meminfo[0] / 1048576,  # Mb
                }
                gpus.append(g)
    except CudaSupportError:
        pass

    return gpus


def clear_memory_all_gpus():
    """Clear memory of all GPUs."""
    try:
        for gpu in cuda.gpus:
            with gpu:
                cuda.current_context().deallocations.clear()
    except CudaSupportError:
        logger.info("No CUDA available")


def get_cuda_version():
    """Get CUDA version

    Returns:
        str: Version of the library.
    """
    try:
        import torch

        return torch.version.cuda
    except (ImportError, ModuleNotFoundError):
        path = ""
        if sys.platform == "win32":
            candidate = (
                "C:\\Program Files\\NVIDIA GPU Computing Toolkit\\CUDA\\v*\\version.txt"
            )
            path_list = glob.glob(candidate)
            if path_list:
                path = path_list[0]
        elif sys.platform == "linux" or sys.platform == "darwin":
            path = "/usr/local/cuda/version.txt"
        else:
            raise ValueError("Not in Windows, Linux or Mac")

        if os.path.isfile(path):
            with open(path, "r") as f:
                data = f.read().replace("\n", "")
            return data
        else:
            return None


def get_cudnn_version():
    """Get the CuDNN version

    Returns:
        str: Version of the library.
    """

    def find_cudnn_in_headers(candiates):
        for c in candidates:
            file = glob.glob(c)
            if file:
                break
        if file:
            with open(file[0], "r") as f:
                version = ""
                for line in f:
                    if "#define CUDNN_MAJOR" in line:
                        version = line.split()[-1]
                    if "#define CUDNN_MINOR" in line:
                        version += "." + line.split()[-1]
                    if "#define CUDNN_PATCHLEVEL" in line:
                        version += "." + line.split()[-1]
            if version:
                return version
            else:
                return None
        else:
            return None

    try:
        import torch

        return str(torch.backends.cudnn.version())
    except (ImportError, ModuleNotFoundError):
        if sys.platform == "win32":
            candidates = [r"C:\NVIDIA\cuda\include\cudnn.h"]
        elif sys.platform == "linux":
            candidates = [
                "/usr/include/cudnn_version.h",
                "/usr/include/x86_64-linux-gnu/cudnn_v[0-99].h",
                "/usr/local/cuda/include/cudnn.h",
                "/usr/include/cudnn.h",
            ]
        elif sys.platform == "darwin":
            candidates = ["/usr/local/cuda/include/cudnn.h", "/usr/include/cudnn.h"]
        else:
            raise ValueError("Not in Windows, Linux or Mac")
        return find_cudnn_in_headers(candidates)



# Copyright (c) Recommenders contributors.
# Licensed under the MIT License.


import pytest
import numpy as np
import pandas as pd
from pandas.testing import assert_frame_equal

from recommenders.evaluation.python_evaluation import (
    precision_at_k,
    recall_at_k,
    ndcg_at_k,
    map_at_k,
    map,
)

try:
    from pyspark.ml.functions import array_to_vector
    from pyspark.sql.types import ArrayType, FloatType, IntegerType, StructField, StructType
    from recommenders.evaluation.spark_evaluation import (
        SparkDiversityEvaluation,
        SparkRankingEvaluation,
        SparkRatingEvaluation,
    )
except ImportError:
    pass  # skip this import if we are in pure python environment


TOL = 0.0001


@pytest.fixture
def spark_data(rating_true, rating_pred, spark):
    df_true = spark.createDataFrame(rating_true)
    df_pred = spark.createDataFrame(rating_pred)

    return df_true, df_pred


@pytest.fixture
def spark_diversity_data(diversity_data, spark):
    train_df, reco_df, item_feature_df = diversity_data
    
    train_df = spark.createDataFrame(train_df)
    reco_df = spark.createDataFrame(reco_df)
    item_feature_df["features"] = item_feature_df["features"].apply(lambda x: x.tolist())
    field = [
        StructField("ItemId", IntegerType(), True),
        StructField("features", ArrayType(FloatType()), True),
    ]
    item_feature_df = spark.createDataFrame(item_feature_df, schema=StructType(field))
    # Array[Float] to VectorUDT
    item_feature_df = item_feature_df.withColumn("features", array_to_vector(item_feature_df["features"]))

    return train_df, reco_df, item_feature_df


@pytest.mark.spark
def test_init_spark(spark):
    assert spark is not None


@pytest.mark.spark
def test_init_spark_rating_eval(spark_data):
    df_true, df_pred = spark_data
    evaluator = SparkRatingEvaluation(df_true, df_pred)
    assert evaluator is not None


@pytest.mark.spark
def test_spark_rmse(spark_data):
    df_true, df_pred = spark_data

    evaluator = SparkRatingEvaluation(df_true, df_true, col_prediction="rating")
    assert evaluator.rmse() == 0

    evaluator = SparkRatingEvaluation(df_true, df_pred)
    assert evaluator.rmse() == pytest.approx(7.254309, TOL)


@pytest.mark.spark
def test_spark_mae(spark_data):
    df_true, df_pred = spark_data

    evaluator = SparkRatingEvaluation(df_true, df_true, col_prediction="rating")
    assert evaluator.mae() == 0

    evaluator = SparkRatingEvaluation(df_true, df_pred)
    assert evaluator.mae() == pytest.approx(6.375, TOL)


@pytest.mark.spark
def test_spark_rsquared(spark_data):
    df_true, df_pred = spark_data

    evaluator = SparkRatingEvaluation(df_true, df_true, col_prediction="rating")
    assert evaluator.rsquared() == pytest.approx(1.0, TOL)

    evaluator = SparkRatingEvaluation(df_true, df_pred)
    assert evaluator.rsquared() == pytest.approx(-31.699029, TOL)


@pytest.mark.spark
def test_spark_exp_var(spark_data):
    df_true, df_pred = spark_data

    evaluator = SparkRatingEvaluation(df_true, df_true, col_prediction="rating")
    assert evaluator.exp_var() == pytest.approx(1.0, TOL)

    evaluator = SparkRatingEvaluation(df_true, df_pred)
    assert evaluator.exp_var() == pytest.approx(-6.4466, TOL)


@pytest.mark.spark
def test_spark_recall_at_k(spark_data):
    df_true, df_pred = spark_data

    evaluator = SparkRankingEvaluation(df_true, df_pred)
    assert evaluator.recall_at_k() == pytest.approx(0.37777, TOL)

    evaluator = SparkRankingEvaluation(
        df_true, df_pred, relevancy_method="by_threshold", threshold=3.5
    )
    assert evaluator.recall_at_k() == pytest.approx(0.37777, TOL)


@pytest.mark.spark
def test_spark_precision_at_k(spark_data, spark):
    df_true, df_pred = spark_data

    evaluator = SparkRankingEvaluation(df_true, df_pred, k=10)
    assert evaluator.precision_at_k() == pytest.approx(0.26666, TOL)

    evaluator = SparkRankingEvaluation(
        df_true, df_pred, relevancy_method="by_threshold", threshold=3.5
    )
    assert evaluator.precision_at_k() == pytest.approx(0.26666, TOL)

    # Check normalization
    single_user = pd.DataFrame(
        {"userID": [1, 1, 1], "itemID": [1, 2, 3], "rating": [5, 4, 3]}
    )
    df_single = spark.createDataFrame(single_user)
    evaluator = SparkRankingEvaluation(
        df_single, df_single, k=3, col_prediction="rating"
    )
    assert evaluator.precision_at_k() == 1

    same_items = pd.DataFrame(
        {
            "userID": [1, 1, 1, 2, 2, 2],
            "itemID": [1, 2, 3, 1, 2, 3],
            "rating": [5, 4, 3, 5, 5, 3],
        }
    )
    df_same = spark.createDataFrame(same_items)
    evaluator = SparkRankingEvaluation(df_same, df_same, k=3, col_prediction="rating")
    assert evaluator.precision_at_k() == 1

    # Check that if the sample size is smaller than k, the maximum precision can not be 1
    # if we do precision@5 when there is only 3 items, we can get a maximum of 3/5.
    evaluator = SparkRankingEvaluation(df_same, df_same, k=5, col_prediction="rating")
    assert evaluator.precision_at_k() == 0.6


@pytest.mark.spark
def test_spark_ndcg_at_k(spark_data):
    df_true, df_pred = spark_data

    evaluator = SparkRankingEvaluation(df_true, df_true, k=10, col_prediction="rating")
    assert evaluator.ndcg_at_k() == 1.0

    evaluator = SparkRankingEvaluation(df_true, df_pred, k=10)
    assert evaluator.ndcg_at_k() == pytest.approx(0.38172, TOL)

    evaluator = SparkRankingEvaluation(
        df_true, df_pred, relevancy_method="by_threshold", threshold=3.5
    )
    assert evaluator.ndcg_at_k() == pytest.approx(0.38172, TOL)


@pytest.mark.spark
def test_spark_map(spark_data):
    df_true, df_pred = spark_data

    evaluator = SparkRankingEvaluation(df_true, df_true, k=10, col_prediction="rating")
    assert evaluator.map() == 1.0

    evaluator = SparkRankingEvaluation(df_true, df_pred, k=10)
    assert evaluator.map() == pytest.approx(0.23613, TOL)

    evaluator = SparkRankingEvaluation(
        df_true, df_pred, relevancy_method="by_threshold", threshold=3.5
    )
    assert evaluator.map() == pytest.approx(0.23613, TOL)


@pytest.mark.spark
def test_spark_map_at_k(spark_data):
    df_true, df_pred = spark_data

    evaluator = SparkRankingEvaluation(df_true, df_true, k=10, col_prediction="rating")
    assert evaluator.map_at_k() == 1.0

    evaluator = SparkRankingEvaluation(df_true, df_pred, k=10)
    assert evaluator.map_at_k() == pytest.approx(0.23613, TOL)

    evaluator = SparkRankingEvaluation(
        df_true, df_pred, relevancy_method="by_threshold", threshold=3.5
    )
    assert evaluator.map_at_k() == pytest.approx(0.23613, TOL)


@pytest.mark.spark
@pytest.mark.parametrize(
    "k,pred_start_row_i,user_id",
    [
        (10, 0, None),
        (3, 0, None),   # Different k
        (10, 0, None),  # Different pred
        (10, 0, 3),     # Test with one user (userID == 3)
    ]
)
def test_spark_python_match(rating_true, rating_pred, spark, k, pred_start_row_i, user_id):
    df_true, df_pred = rating_true, rating_pred
    df_pred = df_pred[pred_start_row_i:]
    if user_id is not None:
        df_pred = df_pred.loc[df_pred["userID"] == 3]
        df_true = df_true.loc[df_true["userID"] == 3]

    dfs_true = spark.createDataFrame(df_true)
    dfs_pred = spark.createDataFrame(df_pred)

    # Test on the original data with k = 10.
    evaluator = SparkRankingEvaluation(dfs_true, dfs_pred, k=k)

    assert recall_at_k(df_true, df_pred, k=k) == pytest.approx(
        evaluator.recall_at_k(), TOL
    )
    assert precision_at_k(df_true, df_pred, k=k) == pytest.approx(
        evaluator.precision_at_k(), TOL
    )
    assert ndcg_at_k(df_true, df_pred, k=k) == pytest.approx(
        evaluator.ndcg_at_k(), TOL
    )
    assert map_at_k(df_true, df_pred, k=k) == pytest.approx(
        evaluator.map_at_k(), TOL
    )
    assert map(df_true, df_pred, k=k) == pytest.approx(
        evaluator.map(), TOL
    )


@pytest.mark.spark
def test_catalog_coverage(spark_diversity_data):
    train_df, reco_df, _ = spark_diversity_data
    evaluator = SparkDiversityEvaluation(
        train_df=train_df, reco_df=reco_df, col_user="UserId", col_item="ItemId"
    )
    c_coverage = evaluator.catalog_coverage()
    assert c_coverage == pytest.approx(0.8, TOL)


@pytest.mark.spark
def test_distributional_coverage(spark_diversity_data):
    train_df, reco_df, _ = spark_diversity_data
    evaluator = SparkDiversityEvaluation(
        train_df=train_df, reco_df=reco_df, col_user="UserId", col_item="ItemId"
    )
    d_coverage = evaluator.distributional_coverage()
    assert d_coverage == pytest.approx(1.9183, TOL)


@pytest.mark.spark
def test_item_novelty(spark_diversity_data):
    train_df, reco_df, _ = spark_diversity_data

    evaluator = SparkDiversityEvaluation(
        train_df=train_df, reco_df=reco_df, col_user="UserId", col_item="ItemId"
    )
    actual = evaluator.historical_item_novelty().toPandas()
    assert_frame_equal(
        pd.DataFrame(
            dict(ItemId=[1, 2, 3, 4, 5], item_novelty=[3.0, 3.0, 2.0, 1.41504, 3.0])
        ),
        actual,
        check_exact=False,
        atol=TOL,
    )
    assert np.all(actual["item_novelty"].values >= 0)
    # Test that novelty is zero when data includes only one item
    train_df_new = train_df.filter("ItemId == 3")
    evaluator = SparkDiversityEvaluation(
        train_df=train_df_new, reco_df=reco_df, col_user="UserId", col_item="ItemId"
    )
    actual = evaluator.historical_item_novelty().toPandas()
    assert actual["item_novelty"].values[0] == 0


@pytest.mark.spark
def test_novelty(spark_diversity_data):
    train_df, reco_df, _ = spark_diversity_data
    evaluator = SparkDiversityEvaluation(
        train_df=train_df, reco_df=reco_df, col_user="UserId", col_item="ItemId"
    )
    assert evaluator.novelty() == pytest.approx(2.83333, TOL)

    # Test that novelty is zero when data includes only one item
    train_df_new = train_df.filter("ItemId == 3")
    reco_df_new = reco_df.filter("ItemId == 3")
    evaluator = SparkDiversityEvaluation(
        train_df=train_df_new, reco_df=reco_df_new, col_user="UserId", col_item="ItemId"
    )
    assert evaluator.novelty() == 0


@pytest.mark.spark
def test_user_diversity(spark_diversity_data):
    train_df, reco_df, _ = spark_diversity_data
    evaluator = SparkDiversityEvaluation(
        train_df=train_df, reco_df=reco_df, col_user="UserId", col_item="ItemId"
    )
    actual = evaluator.user_diversity().toPandas()
    assert_frame_equal(
        pd.DataFrame(
            dict(UserId=[1, 2, 3], user_diversity=[0.29289, 1.0, 0.0])
        ),
        actual,
        check_exact=False,
        atol=TOL,
    )


@pytest.mark.spark
def test_diversity(spark_diversity_data):
    train_df, reco_df, _ = spark_diversity_data
    evaluator = SparkDiversityEvaluation(
        train_df=train_df, reco_df=reco_df, col_user="UserId", col_item="ItemId"
    )
    assert evaluator.diversity() == pytest.approx(0.43096, TOL)


@pytest.mark.spark
def test_user_item_serendipity(spark_diversity_data):
    train_df, reco_df, _ = spark_diversity_data
    evaluator = SparkDiversityEvaluation(
        train_df=train_df,
        reco_df=reco_df,
        col_user="UserId",
        col_item="ItemId",
        col_relevance="Relevance",
    )
    actual = evaluator.user_item_serendipity().toPandas()
    assert_frame_equal(
        pd.DataFrame(
            dict(
                UserId=[1, 1, 2, 2, 3, 3],
                ItemId=[3, 5, 2, 5, 1, 2],
                user_item_serendipity=[
                    0.72783,
                    0.0,
                    0.71132,
                    0.35777,
                    0.80755,
                    0.0,
                ],
            )
        ),
        actual,
        check_exact=False,
        atol=TOL,
    )


@pytest.mark.spark
def test_user_serendipity(spark_diversity_data):
    train_df, reco_df, _ = spark_diversity_data
    evaluator = SparkDiversityEvaluation(
        train_df=train_df,
        reco_df=reco_df,
        col_user="UserId",
        col_item="ItemId",
        col_relevance="Relevance",
    )
    actual = evaluator.user_serendipity().toPandas()
    assert_frame_equal(
        pd.DataFrame(
            dict(UserId=[1, 2, 3], user_serendipity=[0.363915, 0.53455, 0.403775])
        ),
        actual,
        check_exact=False,
        atol=TOL,
    )


@pytest.mark.spark
def test_serendipity(spark_diversity_data):
    train_df, reco_df, _ = spark_diversity_data
    evaluator = SparkDiversityEvaluation(
        train_df=train_df,
        reco_df=reco_df,
        col_user="UserId",
        col_item="ItemId",
        col_relevance="Relevance",
    )
    assert evaluator.serendipity() == pytest.approx(0.43408, TOL)


@pytest.mark.spark
def test_user_diversity_item_feature_vector(spark_diversity_data):
    train_df, reco_df, item_feature_df = spark_diversity_data
    evaluator = SparkDiversityEvaluation(
        train_df=train_df,
        reco_df=reco_df,
        item_feature_df=item_feature_df,
        item_sim_measure="item_feature_vector",
        col_user="UserId",
        col_item="ItemId",
    )
    actual = evaluator.user_diversity().toPandas()
    assert_frame_equal(
        pd.DataFrame(
            dict(UserId=[1, 2, 3], user_diversity=[0.5000, 0.5000, 0.5000])
        ),
        actual,
        check_exact=False,
        atol=TOL,
    )


@pytest.mark.spark
def test_diversity_item_feature_vector(spark_diversity_data):
    train_df, reco_df, item_feature_df = spark_diversity_data
    evaluator = SparkDiversityEvaluation(
        train_df=train_df,
        reco_df=reco_df,
        item_feature_df=item_feature_df,
        item_sim_measure="item_feature_vector",
        col_user="UserId",
        col_item="ItemId",
    )
    assert evaluator.diversity() == pytest.approx(0.5000, TOL)


@pytest.mark.spark
def test_user_item_serendipity_item_feature_vector(
    spark_diversity_data,
):
    train_df, reco_df, item_feature_df = spark_diversity_data
    evaluator = SparkDiversityEvaluation(
        train_df=train_df,
        reco_df=reco_df,
        item_feature_df=item_feature_df,
        item_sim_measure="item_feature_vector",
        col_user="UserId",
        col_item="ItemId",
        col_relevance="Relevance",
    )
    actual = evaluator.user_item_serendipity().toPandas()
    assert_frame_equal(
        pd.DataFrame(
            dict(
                UserId=[1, 1, 2, 2, 3, 3],
                ItemId=[3, 5, 2, 5, 1, 2],
                user_item_serendipity=[
                    0.5000,
                    0.0,
                    0.75,
                    0.5000,
                    0.6667,
                    0.0,
                ],
            )
        ),
        actual,
        check_exact=False,
        atol=TOL,
    )


@pytest.mark.spark
def test_user_serendipity_item_feature_vector(spark_diversity_data):
    train_df, reco_df, item_feature_df = spark_diversity_data
    evaluator = SparkDiversityEvaluation(
        train_df=train_df,
        reco_df=reco_df,
        item_feature_df=item_feature_df,
        item_sim_measure="item_feature_vector",
        col_user="UserId",
        col_item="ItemId",
        col_relevance="Relevance",
    )
    actual = evaluator.user_serendipity().toPandas()
    assert_frame_equal(
        pd.DataFrame(
            dict(UserId=[1, 2, 3], user_serendipity=[0.2500, 0.625, 0.3333])
        ),
        actual,
        check_exact=False,
        atol=TOL,
    )


@pytest.mark.spark
def test_serendipity_item_feature_vector(spark_diversity_data):
    train_df, reco_df, item_feature_df = spark_diversity_data
    evaluator = SparkDiversityEvaluation(
        train_df=train_df,
        reco_df=reco_df,
        item_feature_df=item_feature_df,
        item_sim_measure="item_feature_vector",
        col_user="UserId",
        col_item="ItemId",
        col_relevance="Relevance",
    )
    assert evaluator.serendipity() == pytest.approx(0.4028, TOL)



# Copyright (c) Recommenders contributors.
# Licensed under the MIT License.

# NOTE:
# The times on GPU environment have been calculated on an Azure STANDARD_NC6S_V2
# with 6 vCPUs, 112 GB memory, 1 NVIDIA Tesla P100 GPU.
# The times on CPU and Spark environments have been calculated on an Azure
# Standard_A8m_v2 with 8 vCPUs and 64 GiB memory.

# IMPORTANT NOTE:
# FOR NIGHTLY, NO GROUP SHOULD SURPASS 45MIN = 2700s !!!
# FOR PR GATE, NO GROUP SHOULD SURPASS 15MIN = 900s !!!

global nightly_test_groups, pr_gate_test_groups

nightly_test_groups = {
    "group_cpu_001": [  # Total group time: 1883s
        "tests/data_validation/recommenders/datasets/test_movielens.py::test_download_and_extract_movielens",  # 0.45s + 0.61s + 3.47s + 8.28s
        "tests/data_validation/recommenders/datasets/test_movielens.py::test_load_item_df",  # 0.47s + 0.59s + 3.59s + 8.44s
        "tests/data_validation/recommenders/datasets/test_movielens.py::test_load_pandas_df",  # 16.87s + 37.33s + 352.99s + 673.61s
        #
        "tests/data_validation/recommenders/datasets/test_mind.py::test_mind_url",  # 0.38s
        "tests/data_validation/recommenders/datasets/test_mind.py::test_download_mind_demo",
        "tests/data_validation/recommenders/datasets/test_mind.py::test_extract_mind_demo",
        "tests/data_validation/recommenders/datasets/test_mind.py::test_download_mind_small",
        "tests/data_validation/recommenders/datasets/test_mind.py::test_extract_mind_small",
        "tests/data_validation/recommenders/datasets/test_mind.py::test_download_mind_large",
        "tests/data_validation/recommenders/datasets/test_mind.py::test_extract_mind_large",
        "tests/data_validation/examples/test_mind.py::test_mind_utils_runs",  # 219.77s
        "tests/data_validation/examples/test_mind.py::test_mind_utils_values",  # 219.26s
        #
        "tests/data_validation/examples/test_wikidata.py::test_wikidata_runs",
        "tests/data_validation/examples/test_wikidata.py::test_wikidata_values",
        #
        "tests/smoke/examples/test_notebooks_python.py::test_lightgbm_quickstart_smoke",  # 46.42s
        #
        "tests/smoke/examples/test_notebooks_python.py::test_cornac_bpr_smoke",  # 16.62s
        "tests/functional/examples/test_notebooks_python.py::test_cornac_bpr_functional",  # 165.72s
    ],
    "group_cpu_002": [  # Total group time: 1801s
        "tests/smoke/examples/test_notebooks_python.py::test_baseline_deep_dive_smoke",  # 15.98s
        "tests/functional/examples/test_notebooks_python.py::test_baseline_deep_dive_functional",  # 170.73s
        #
        "tests/smoke/examples/test_notebooks_python.py::test_surprise_svd_smoke",  # 45.88s
        "tests/functional/examples/test_notebooks_python.py::test_surprise_svd_functional",  # 503.54s
        #
        "tests/functional/examples/test_notebooks_python.py::test_geoimc_functional",  # 1006.19s
        #
        "tests/functional/examples/test_notebooks_python.py::test_benchmark_movielens_cpu",  # 58s
    ],
    "group_cpu_003": [  # Total group time: 2253s
        "tests/data_validation/recommenders/datasets/test_criteo.py::test_download_criteo_sample",  # 1.05s
        "tests/data_validation/recommenders/datasets/test_criteo.py::test_extract_criteo_sample",  # 1.22s
        "tests/data_validation/recommenders/datasets/test_criteo.py::test_criteo_load_pandas_df_sample",  # 1.73s
        "tests/data_validation/recommenders/datasets/test_criteo.py::test_criteo_load_pandas_df_full",  # 1368.63s
        #
        "tests/smoke/examples/test_notebooks_python.py::test_sar_single_node_smoke",  # 12.58s
        "tests/functional/examples/test_notebooks_python.py::test_sar_single_node_functional",  # 57.67s + 808.83s
        "tests/functional/examples/test_notebooks_python.py::test_xlearn_fm_functional",  # 255.73s
        "tests/smoke/examples/test_notebooks_python.py::test_vw_deep_dive_smoke",
        "tests/functional/examples/test_notebooks_python.py::test_vw_deep_dive_functional",
        "tests/functional/examples/test_notebooks_python.py::test_nni_tuning_svd",
    ],
    "group_gpu_001": [  # Total group time: 1937.01s
        "tests/unit/examples/test_notebooks_gpu.py::test_gpu_vm",  # 0.76s (Always the first test to check the GPU works)
        "tests/smoke/recommenders/models/test_deeprec_utils.py",  # 2.91
        "tests/smoke/recommenders/models/test_deeprec_model.py::test_FFM_iterator",  # 0.74s
        "tests/smoke/recommenders/models/test_newsrec_utils.py::test_news_iterator",  # 3.04s
        #
        "tests/smoke/recommenders/models/test_deeprec_model.py::test_model_lightgcn",  # 6.03s
        "tests/functional/examples/test_notebooks_gpu.py::test_lightgcn_deep_dive_functional",  # 19.45s
        #
        # "tests/smoke/recommenders/models/test_deeprec_model.py::test_model_sum",  # 27.23s  # FIXME: Disabled due to the issue with TF version > 2.10.1 See #2018
        #
        "tests/smoke/recommenders/models/test_deeprec_model.py::test_model_dkn",  # 187.20s
        "tests/functional/examples/test_notebooks_gpu.py::test_dkn_quickstart_functional",  # 1167.93s
        #
        "tests/functional/examples/test_notebooks_gpu.py::test_slirec_quickstart_functional",  # 175.00s
        "tests/smoke/recommenders/models/test_deeprec_model.py::test_model_slirec",  # 346.72s
    ],
    "group_gpu_002": [  # Total group time: 1896.76s
        "tests/unit/examples/test_notebooks_gpu.py::test_gpu_vm",  # 0.76s (Always the first test to check the GPU works)
        "tests/smoke/recommenders/models/test_deeprec_model.py::test_model_xdeepfm",  # 3.10s
        # FIXME: https://github.com/microsoft/recommenders/issues/1883
        # "tests/smoke/examples/test_notebooks_gpu.py::test_xdeepfm_smoke",  # 77.93s
        "tests/functional/examples/test_notebooks_gpu.py::test_xdeepfm_functional",
        #
        "tests/smoke/examples/test_notebooks_gpu.py::test_cornac_bivae_smoke",  # 67.84s
        "tests/functional/examples/test_notebooks_gpu.py::test_cornac_bivae_functional",  # 453.21s
        #
        "tests/smoke/examples/test_notebooks_gpu.py::test_wide_deep_smoke",  # 122.71s
        #
        "tests/smoke/examples/test_notebooks_gpu.py::test_fastai_smoke",  # 33.22s
        "tests/functional/examples/test_notebooks_gpu.py::test_fastai_functional",  # 667.88s
    ],
    "group_gpu_003": [  # Total group time: 2072.15s
        "tests/unit/examples/test_notebooks_gpu.py::test_gpu_vm",  # 0.76s (Always the first test to check the GPU works)
        "tests/smoke/examples/test_notebooks_gpu.py::test_ncf_smoke",  # 114.39s
        "tests/functional/examples/test_notebooks_gpu.py::test_ncf_functional",  # 1046.97s
        "tests/smoke/examples/test_notebooks_gpu.py::test_ncf_deep_dive_smoke",  # 102.71s
        "tests/functional/examples/test_notebooks_gpu.py::test_ncf_deep_dive_functional",  # 351.17s
        #
        "tests/smoke/recommenders/models/test_newsrec_utils.py::test_naml_iterator",  # 5.50s
        # FIXME: https://github.com/microsoft/recommenders/issues/1883
        # "tests/smoke/recommenders/models/test_newsrec_model.py::test_model_naml",  # 450.65s
    ],
    "group_gpu_004": [  # Total group time: 2103.34s
        "tests/unit/examples/test_notebooks_gpu.py::test_gpu_vm",  # 0.76s (Always the first test to check the GPU works)
        "tests/smoke/examples/test_notebooks_gpu.py::test_nrms_smoke",  # 232.55s
        # FIXME: https://github.com/microsoft/recommenders/issues/1883
        # "tests/functional/examples/test_notebooks_gpu.py::test_nrms_quickstart_functional",  # 857.05s
        #
        "tests/smoke/examples/test_notebooks_gpu.py::test_lstur_smoke",  # 246.46s
        # FIXME: https://github.com/microsoft/recommenders/issues/1883
        # "tests/functional/examples/test_notebooks_gpu.py::test_lstur_quickstart_functional",  # 766.52s
    ],
    "group_gpu_005": [  # Total group time: 1844.05s
        "tests/unit/examples/test_notebooks_gpu.py::test_gpu_vm",  # 0.76s (Always the first test to check the GPU works)
        # FIXME: https://github.com/microsoft/recommenders/issues/1883
        # "tests/functional/examples/test_notebooks_gpu.py::test_wide_deep_functional",  # 1843.29s
        #
        "tests/smoke/examples/test_notebooks_gpu.py::test_npa_smoke",  # 366.22s
        # FIXME: https://github.com/microsoft/recommenders/issues/1883
        # "tests/functional/examples/test_notebooks_gpu.py::test_npa_quickstart_functional",  # 810.92s
    ],
    "group_gpu_006": [  # Total group time: 1763.99s
        "tests/unit/examples/test_notebooks_gpu.py::test_gpu_vm",  # 0.76s (Always the first test to check the GPU works)
        "tests/smoke/recommenders/models/test_newsrec_model.py::test_model_npa",  # 202.61s
        "tests/smoke/recommenders/models/test_newsrec_model.py::test_model_nrms",  # 188.60s
    ],
    "group_gpu_007": [  # Total group time: 846.89s
        "tests/unit/examples/test_notebooks_gpu.py::test_gpu_vm",  # 0.76s (Always the first test to check the GPU works)
        # FIXME: https://github.com/microsoft/recommenders/issues/1883
        # "tests/smoke/examples/test_notebooks_gpu.py::test_naml_smoke",  # 620.13s
        #
        "tests/functional/examples/test_notebooks_gpu.py::test_benchmark_movielens_gpu",  # 226s
        # FIXME: Reduce test time https://github.com/microsoft/recommenders/issues/1731
        # "tests/functional/examples/test_notebooks_gpu.py::test_naml_quickstart_functional",  # 2033.85s
        # FIXME: https://github.com/microsoft/recommenders/issues/1716
        # "tests/functional/examples/test_notebooks_gpu.py::test_sasrec_quickstart_functional",  # 448.06s + 614.69s
        "tests/smoke/recommenders/models/test_newsrec_model.py::test_model_lstur",  # 194.88s
    ],
    "group_spark_001": [  # Total group time: 987.16s
        "tests/data_validation/recommenders/datasets/test_movielens.py::test_load_spark_df",  # 4.33s+ 25.58s + 101.99s + 139.23s
        #
        "tests/data_validation/recommenders/datasets/test_criteo.py::test_criteo_load_spark_df_sample",  # 6.83s
        "tests/data_validation/recommenders/datasets/test_criteo.py::test_criteo_load_spark_df_full",  # 374.64s
        #
        "tests/smoke/examples/test_notebooks_pyspark.py::test_mmlspark_lightgbm_criteo_smoke",  # 32.45s
        "tests/functional/examples/test_notebooks_pyspark.py::test_mmlspark_lightgbm_criteo_functional",
        #
        "tests/smoke/examples/test_notebooks_pyspark.py::test_als_pyspark_smoke",  # 49.53s
        "tests/functional/examples/test_notebooks_pyspark.py::test_als_pyspark_functional",  # 110.58s
        #
        "tests/functional/examples/test_notebooks_pyspark.py::test_benchmark_movielens_pyspark",  # 142s
    ],
}

pr_gate_test_groups = {
    "group_cpu_001": [  # Total group time: 525.96s
        "tests/data_validation/recommenders/datasets/test_movielens.py::test_mock_movielens_schema__has_default_col_names",
        "tests/data_validation/recommenders/datasets/test_movielens.py::test_mock_movielens_schema__get_df_remove_default_col__return_success",
        "tests/data_validation/recommenders/datasets/test_movielens.py::test_mock_movielens_schema__get_df_invalid_param__return_failure",
        "tests/data_validation/recommenders/datasets/test_movielens.py::test_mock_movielens_schema__get_df__return_success",
        "tests/data_validation/recommenders/datasets/test_movielens.py::test_mock_movielens_data__no_name_collision",
        "tests/data_validation/recommenders/datasets/test_movielens.py::test_load_pandas_df_mock_100__with_default_param__succeed",
        "tests/data_validation/recommenders/datasets/test_movielens.py::test_load_pandas_df_mock_100__with_custom_param__succeed",
        "tests/data_validation/recommenders/datasets/test_wikidata.py::test_find_wikidata_id",
        "tests/unit/recommenders/datasets/test_download_utils.py::test_maybe_download",
        "tests/unit/recommenders/datasets/test_download_utils.py::test_maybe_download_wrong_bytes",
        "tests/unit/recommenders/datasets/test_download_utils.py::test_maybe_download_maybe",
        "tests/unit/recommenders/datasets/test_download_utils.py::test_maybe_download_retry",
        "tests/unit/recommenders/datasets/test_download_utils.py::test_download_path",
        "tests/unit/recommenders/datasets/test_pandas_df_utils.py::test_negative_feedback_sampler",
        "tests/unit/recommenders/datasets/test_pandas_df_utils.py::test_filter_by",
        "tests/unit/recommenders/datasets/test_pandas_df_utils.py::test_csv_to_libffm",
        "tests/unit/recommenders/datasets/test_pandas_df_utils.py::test_has_columns",
        "tests/unit/recommenders/datasets/test_pandas_df_utils.py::test_has_same_base_dtype",
        "tests/unit/recommenders/datasets/test_pandas_df_utils.py::test_lru_cache_df",
        "tests/unit/recommenders/datasets/test_python_splitter.py::test_split_pandas_data",
        "tests/unit/recommenders/datasets/test_python_splitter.py::test_min_rating_filter",
        "tests/unit/recommenders/datasets/test_python_splitter.py::test_random_splitter",
        "tests/unit/recommenders/datasets/test_python_splitter.py::test_chrono_splitter",
        "tests/unit/recommenders/datasets/test_python_splitter.py::test_stratified_splitter",
        "tests/unit/recommenders/datasets/test_python_splitter.py::test_int_numpy_stratified_splitter",
        "tests/unit/recommenders/datasets/test_python_splitter.py::test_float_numpy_stratified_splitter",
        "tests/unit/recommenders/datasets/test_sparse.py::test_df_to_sparse",
        "tests/unit/recommenders/datasets/test_sparse.py::test_sparse_to_df",
        "tests/unit/recommenders/evaluation/test_python_evaluation.py::test_column_dtypes_match",
        "tests/unit/recommenders/evaluation/test_python_evaluation.py::test_merge_rating",
        "tests/unit/recommenders/evaluation/test_python_evaluation.py::test_merge_ranking",
        "tests/unit/recommenders/evaluation/test_python_evaluation.py::test_python_rmse",
        "tests/unit/recommenders/evaluation/test_python_evaluation.py::test_python_mae",
        "tests/unit/recommenders/evaluation/test_python_evaluation.py::test_python_rsquared",
        "tests/unit/recommenders/evaluation/test_python_evaluation.py::test_python_exp_var",
        "tests/unit/recommenders/evaluation/test_python_evaluation.py::test_get_top_k_items",
        "tests/unit/recommenders/evaluation/test_python_evaluation.py::test_get_top_k_items_largek",
        "tests/unit/recommenders/evaluation/test_python_evaluation.py::test_python_ndcg_at_k",
        "tests/unit/recommenders/evaluation/test_python_evaluation.py::test_python_map",
        "tests/unit/recommenders/evaluation/test_python_evaluation.py::test_python_map_at_k",
        "tests/unit/recommenders/evaluation/test_python_evaluation.py::test_python_precision_at_k",
        "tests/unit/recommenders/evaluation/test_python_evaluation.py::test_python_recall_at_k",
        "tests/unit/recommenders/evaluation/test_python_evaluation.py::test_python_auc",
        "tests/unit/recommenders/evaluation/test_python_evaluation.py::test_python_logloss",
        "tests/unit/recommenders/evaluation/test_python_evaluation.py::test_python_errors",
        "tests/unit/recommenders/evaluation/test_python_evaluation.py::test_catalog_coverage",
        "tests/unit/recommenders/evaluation/test_python_evaluation.py::test_distributional_coverage",
        "tests/unit/recommenders/evaluation/test_python_evaluation.py::test_item_novelty",
        "tests/unit/recommenders/evaluation/test_python_evaluation.py::test_novelty",
        "tests/unit/recommenders/evaluation/test_python_evaluation.py::test_user_diversity",
        "tests/unit/recommenders/evaluation/test_python_evaluation.py::test_diversity",
        "tests/unit/recommenders/evaluation/test_python_evaluation.py::test_user_item_serendipity",
        "tests/unit/recommenders/evaluation/test_python_evaluation.py::test_user_serendipity",
        "tests/unit/recommenders/evaluation/test_python_evaluation.py::test_serendipity",
        "tests/unit/recommenders/evaluation/test_python_evaluation.py::test_user_diversity_item_feature_vector",
        "tests/unit/recommenders/evaluation/test_python_evaluation.py::test_diversity_item_feature_vector",
        "tests/unit/recommenders/evaluation/test_python_evaluation.py::test_user_item_serendipity_item_feature_vector",
        "tests/unit/recommenders/evaluation/test_python_evaluation.py::test_user_serendipity_item_feature_vector",
        "tests/unit/recommenders/evaluation/test_python_evaluation.py::test_serendipity_item_feature_vector",
        "tests/unit/recommenders/models/test_tfidf_utils.py::test_init",
        "tests/unit/recommenders/models/test_tfidf_utils.py::test_clean_dataframe",
        "tests/unit/recommenders/models/test_tfidf_utils.py::test_fit",
        "tests/unit/recommenders/models/test_tfidf_utils.py::test_tokenize_text",
        "tests/unit/recommenders/models/test_tfidf_utils.py::test_get_tokens",
        "tests/unit/recommenders/models/test_tfidf_utils.py::test_get_stop_words",
        "tests/unit/recommenders/models/test_tfidf_utils.py::test_recommend_top_k_items",
        "tests/unit/recommenders/models/test_tfidf_utils.py::test_get_top_k_recommendations",
        "tests/unit/recommenders/models/test_cornac_utils.py::test_predict",
        "tests/unit/recommenders/models/test_cornac_utils.py::test_recommend_k_items",
        "tests/unit/recommenders/models/test_sar_singlenode.py::test_init",
        "tests/unit/recommenders/models/test_sar_singlenode.py::test_fit",
        "tests/unit/recommenders/models/test_sar_singlenode.py::test_predict",
        "tests/unit/recommenders/models/test_sar_singlenode.py::test_predict_all_items",
        "tests/unit/recommenders/models/test_sar_singlenode.py::test_sar_item_similarity",
        "tests/unit/recommenders/models/test_sar_singlenode.py::test_user_affinity",
        "tests/unit/recommenders/models/test_sar_singlenode.py::test_recommend_k_items",
        "tests/unit/recommenders/models/test_sar_singlenode.py::test_get_item_based_topk",
        "tests/unit/recommenders/models/test_sar_singlenode.py::test_get_popularity_based_topk",
        "tests/unit/recommenders/models/test_sar_singlenode.py::test_get_normalized_scores",
        "tests/unit/recommenders/models/test_sar_singlenode.py::test_match_similarity_type_from_json_file",
        "tests/unit/recommenders/models/test_sar_singlenode.py::test_dataset_with_duplicates",
        "tests/unit/recommenders/models/test_sar_singlenode.py::test_get_topk_most_similar_users",
        "tests/unit/recommenders/models/test_sar_singlenode.py::test_item_frequencies",
        "tests/unit/recommenders/models/test_sar_singlenode.py::test_user_frequencies",
        "tests/unit/recommenders/models/test_surprise_utils.py::test_predict",
        "tests/unit/recommenders/models/test_surprise_utils.py::test_recommend_k_items",
        "tests/unit/recommenders/models/test_vowpal_wabbit.py::test_vw_init_del",
        "tests/unit/recommenders/models/test_vowpal_wabbit.py::test_to_vw_cmd",
        "tests/unit/recommenders/models/test_vowpal_wabbit.py::test_parse_train_cmd",
        "tests/unit/recommenders/models/test_vowpal_wabbit.py::test_parse_test_cmd",
        "tests/unit/recommenders/models/test_vowpal_wabbit.py::test_to_vw_file",
        "tests/unit/recommenders/models/test_vowpal_wabbit.py::test_fit_and_predict",
        "tests/unit/recommenders/tuning/test_ncf_utils.py::test_compute_test_results__return_success",
        "tests/unit/recommenders/tuning/test_nni_utils.py::test_get_experiment_status",
        "tests/unit/recommenders/tuning/test_nni_utils.py::test_check_experiment_status_done",
        "tests/unit/recommenders/tuning/test_nni_utils.py::test_check_experiment_status_tuner_no_more_trial",
        "tests/unit/recommenders/tuning/test_nni_utils.py::test_check_experiment_status_running",
        "tests/unit/recommenders/tuning/test_nni_utils.py::test_check_experiment_status_no_more_trial",
        "tests/unit/recommenders/tuning/test_nni_utils.py::test_check_experiment_status_failed",
        "tests/unit/recommenders/tuning/test_nni_utils.py::test_check_stopped_timeout",
        "tests/unit/recommenders/tuning/test_nni_utils.py::test_check_stopped",
        "tests/unit/recommenders/tuning/test_nni_utils.py::test_check_metrics_written",
        "tests/unit/recommenders/tuning/test_nni_utils.py::test_check_metrics_written_timeout",
        "tests/unit/recommenders/tuning/test_nni_utils.py::test_get_trials",
        "tests/unit/recommenders/tuning/test_sweep.py::test_param_sweep",
        "tests/unit/recommenders/utils/test_general_utils.py::test_invert_dictionary",
        "tests/unit/recommenders/utils/test_general_utils.py::test_get_number_processors",
        "tests/unit/recommenders/utils/test_plot.py::test_line_graph",
        "tests/unit/recommenders/utils/test_python_utils.py::test_python_jaccard",
        "tests/unit/recommenders/utils/test_python_utils.py::test_python_lift",
        "tests/unit/recommenders/utils/test_python_utils.py::test_exponential_decay",
        "tests/unit/recommenders/utils/test_python_utils.py::test_get_top_k_scored_items",
        "tests/unit/recommenders/utils/test_python_utils.py::test_binarize",
        "tests/unit/recommenders/utils/test_python_utils.py::test_rescale",
        "tests/unit/recommenders/utils/test_timer.py::test_no_time",
        "tests/unit/recommenders/utils/test_timer.py::test_stop_before_start",
        "tests/unit/recommenders/utils/test_timer.py::test_interval_before_stop",
        "tests/unit/recommenders/utils/test_timer.py::test_timer",
        "tests/unit/recommenders/utils/test_timer.py::test_timer_format",
        "tests/performance/recommenders/evaluation/test_python_evaluation_time_performance.py::test_merge_rating",
        "tests/performance/recommenders/evaluation/test_python_evaluation_time_performance.py::test_merge_ranking",
        "tests/performance/recommenders/evaluation/test_python_evaluation_time_performance.py::test_python_rmse",
        "tests/performance/recommenders/evaluation/test_python_evaluation_time_performance.py::test_python_mae",
        "tests/performance/recommenders/evaluation/test_python_evaluation_time_performance.py::test_python_rsquared",
        "tests/performance/recommenders/evaluation/test_python_evaluation_time_performance.py::test_python_exp_var",
        "tests/performance/recommenders/evaluation/test_python_evaluation_time_performance.py::test_get_top_k_items",
        "tests/performance/recommenders/evaluation/test_python_evaluation_time_performance.py::test_get_top_k_items_largek",
        "tests/performance/recommenders/evaluation/test_python_evaluation_time_performance.py::test_python_ndcg_at_k",
        "tests/performance/recommenders/evaluation/test_python_evaluation_time_performance.py::test_python_map_at_k",
        "tests/performance/recommenders/evaluation/test_python_evaluation_time_performance.py::test_python_precision",
        "tests/performance/recommenders/evaluation/test_python_evaluation_time_performance.py::test_python_recall",
        "tests/performance/recommenders/evaluation/test_python_evaluation_time_performance.py::test_python_auc",
        "tests/performance/recommenders/evaluation/test_python_evaluation_time_performance.py::test_python_logloss",
        "tests/security/test_dependency_security.py::test_requests",
        "tests/security/test_dependency_security.py::test_numpy",
        "tests/security/test_dependency_security.py::test_pandas",
        "tests/responsible_ai/recommenders/datasets/test_criteo_privacy.py",
        "tests/responsible_ai/recommenders/datasets/test_movielens_privacy.py",
        "tests/integration/recommenders/utils/test_k8s_utils.py",
    ],
    "group_notebooks_cpu_001": [  # Total group time: 226.42s
        "tests/unit/examples/test_notebooks_python.py::test_sar_deep_dive_runs",
        "tests/unit/examples/test_notebooks_python.py::test_baseline_deep_dive_runs",
        "tests/unit/examples/test_notebooks_python.py::test_template_runs",
        "tests/unit/recommenders/utils/test_notebook_utils.py::test_is_jupyter",
        "tests/unit/recommenders/utils/test_notebook_utils.py::test_update_parameters",
        "tests/unit/recommenders/utils/test_notebook_utils.py::test_notebook_execution",
        "tests/unit/recommenders/utils/test_notebook_utils.py::test_notebook_execution_with_parameters",
        "tests/unit/recommenders/utils/test_notebook_utils.py::test_notebook_execution_value_error_fails",
        "tests/unit/examples/test_notebooks_python.py::test_surprise_deep_dive_runs",
        "tests/unit/examples/test_notebooks_python.py::test_lightgbm",
        "tests/unit/examples/test_notebooks_python.py::test_cornac_deep_dive_runs",
        "tests/unit/examples/test_notebooks_python.py::test_sar_single_node_runs",
        "tests/unit/examples/test_notebooks_python.py::test_vw_deep_dive_runs",
    ],
    "group_spark_001": [  # Total group time: 270.41s
        "tests/unit/recommenders/evaluation/test_spark_evaluation.py::test_init_spark",
        "tests/data_validation/recommenders/datasets/test_movielens.py::test_mock_movielens_schema__get_spark_df__return_success",
        "tests/data_validation/recommenders/datasets/test_movielens.py::test_mock_movielens_schema__get_spark_df__store_tmp_file",
        "tests/data_validation/recommenders/datasets/test_movielens.py::test_mock_movielens_schema__get_spark_df__data_serialization_default_param",
        "tests/data_validation/recommenders/datasets/test_movielens.py::test_load_spark_df_mock_100__with_default_param__succeed",
        "tests/data_validation/recommenders/datasets/test_movielens.py::test_load_spark_df_mock_100__with_custom_param__succeed",
        "tests/unit/recommenders/datasets/test_spark_splitter.py::test_stratified_splitter",
        "tests/unit/recommenders/datasets/test_spark_splitter.py::test_chrono_splitter",
        "tests/unit/recommenders/evaluation/test_spark_evaluation.py::test_user_diversity_item_feature_vector",
        "tests/unit/recommenders/evaluation/test_spark_evaluation.py::test_spark_python_match",
        "tests/unit/recommenders/evaluation/test_spark_evaluation.py::test_spark_exp_var",
        "tests/unit/recommenders/evaluation/test_spark_evaluation.py::test_user_diversity",
        "tests/unit/recommenders/evaluation/test_spark_evaluation.py::test_user_item_serendipity",
        "tests/unit/recommenders/datasets/test_spark_splitter.py::test_min_rating_filter",
        "tests/unit/recommenders/evaluation/test_spark_evaluation.py::test_serendipity",
        "tests/unit/recommenders/evaluation/test_spark_evaluation.py::test_user_serendipity",
        "tests/unit/recommenders/evaluation/test_spark_evaluation.py::test_diversity_item_feature_vector",
        "tests/unit/recommenders/evaluation/test_spark_evaluation.py::test_diversity",
        "tests/unit/recommenders/evaluation/test_spark_evaluation.py::test_user_serendipity_item_feature_vector",
        "tests/unit/recommenders/evaluation/test_spark_evaluation.py::test_serendipity_item_feature_vector",
        "tests/unit/recommenders/evaluation/test_spark_evaluation.py::test_user_item_serendipity_item_feature_vector",
        "tests/unit/recommenders/evaluation/test_spark_evaluation.py::test_novelty",
        "tests/unit/recommenders/datasets/test_spark_splitter.py::test_timestamp_splitter",
        "tests/unit/recommenders/evaluation/test_spark_evaluation.py::test_spark_ndcg_at_k",
        "tests/unit/recommenders/evaluation/test_spark_evaluation.py::test_spark_recall_at_k",
        "tests/unit/recommenders/evaluation/test_spark_evaluation.py::test_spark_precision_at_k",
        "tests/unit/recommenders/evaluation/test_spark_evaluation.py::test_spark_map_at_k",
        "tests/unit/recommenders/evaluation/test_spark_evaluation.py::test_spark_map",
        "tests/unit/recommenders/evaluation/test_spark_evaluation.py::test_spark_rmse",
        "tests/unit/recommenders/evaluation/test_spark_evaluation.py::test_item_novelty",
        "tests/unit/recommenders/evaluation/test_spark_evaluation.py::test_spark_mae",
        "tests/unit/recommenders/evaluation/test_spark_evaluation.py::test_spark_rsquared",
        "tests/unit/recommenders/datasets/test_spark_splitter.py::test_random_splitter",
        "tests/unit/recommenders/evaluation/test_spark_evaluation.py::test_init_spark_rating_eval",
        "tests/unit/recommenders/evaluation/test_spark_evaluation.py::test_catalog_coverage",
        "tests/unit/recommenders/evaluation/test_spark_evaluation.py::test_distributional_coverage",
    ],
    "group_notebooks_spark_001": [  # Total group time: 794s
        "tests/unit/recommenders/utils/test_notebook_utils.py::test_is_databricks",
        "tests/unit/examples/test_notebooks_pyspark.py::test_als_deep_dive_runs",  # 287.70s
        "tests/unit/examples/test_notebooks_pyspark.py::test_als_pyspark_runs",  # 374.15s
        "tests/unit/examples/test_notebooks_pyspark.py::test_mmlspark_lightgbm_criteo_runs",  # 132.09s
    ],
    "group_notebooks_spark_002": [  # Total group time: 466s
        "tests/unit/examples/test_notebooks_pyspark.py::test_data_split_runs",  # 43.66s
        "tests/unit/examples/test_notebooks_pyspark.py::test_evaluation_runs",  # 45.24s
        "tests/unit/examples/test_notebooks_pyspark.py::test_evaluation_diversity_runs",  # 376.61s
        # TODO: This is a flaky test, skip for now, to be fixed in future iterations.
        # Refer to the issue: https://github.com/microsoft/recommenders/issues/1770
        # "tests/unit/examples/test_notebooks_pyspark.py::test_spark_tuning",  # 212.29s+190.02s+180.13s+164.09s=746.53s (flaky test, it rerun several times)
    ],
    "group_gpu_001": [  # Total group time: 492.62s
        "tests/unit/examples/test_notebooks_gpu.py::test_gpu_vm",  # 0.76s (Always the first test to check the GPU works)
        "tests/unit/recommenders/models/test_rbm.py::test_class_init",
        "tests/unit/recommenders/models/test_rbm.py::test_sampling_funct",
        "tests/unit/recommenders/models/test_rbm.py::test_train_param_init",
        "tests/unit/recommenders/models/test_rbm.py::test_save_load",
        "tests/unit/recommenders/models/test_ncf_dataset.py::test_datafile_init",
        "tests/unit/recommenders/models/test_ncf_dataset.py::test_train_loader",
        "tests/unit/recommenders/models/test_ncf_dataset.py::test_test_loader",
        "tests/unit/recommenders/models/test_ncf_dataset.py::test_datafile_init_unsorted",
        "tests/unit/recommenders/models/test_ncf_dataset.py::test_datafile_init_empty",
        "tests/unit/recommenders/models/test_ncf_dataset.py::test_datafile_missing_column",
        "tests/unit/recommenders/models/test_ncf_dataset.py::test_negative_sampler",
        "tests/unit/recommenders/models/test_ncf_singlenode.py::test_init",
        "tests/unit/recommenders/models/test_ncf_singlenode.py::test_fit",
        "tests/unit/recommenders/models/test_ncf_singlenode.py::test_neumf_save_load",
        "tests/unit/recommenders/models/test_ncf_singlenode.py::test_regular_save_load",
        "tests/unit/recommenders/models/test_ncf_singlenode.py::test_predict",
        "tests/unit/recommenders/models/test_wide_deep_utils.py::test_wide_model",
        "tests/unit/recommenders/models/test_wide_deep_utils.py::test_deep_model",
        "tests/unit/recommenders/models/test_wide_deep_utils.py::test_wide_deep_model",
        "tests/unit/recommenders/models/test_newsrec_model.py::test_naml_component_definition",
        "tests/unit/recommenders/models/test_newsrec_model.py::test_lstur_component_definition",
        "tests/unit/recommenders/models/test_newsrec_model.py::test_nrms_component_definition",
        "tests/unit/recommenders/models/test_newsrec_model.py::test_npa_component_definition",
        "tests/unit/recommenders/models/test_newsrec_utils.py::test_prepare_hparams",
        "tests/unit/recommenders/models/test_newsrec_utils.py::test_load_yaml_file",
        # "tests/unit/recommenders/models/test_sasrec_model.py::test_prepare_data", # FIXME: it takes too long to run
        # "tests/unit/recommenders/models/test_sasrec_model.py::test_sampler", # FIXME: it takes too long to run
        # "tests/unit/recommenders/models/test_sasrec_model.py::test_sasrec", # FIXME: it takes too long to run
        # "tests/unit/recommenders/models/test_sasrec_model.py::test_ssept", # FIXME: it takes too long to run
        "tests/unit/recommenders/utils/test_gpu_utils.py::test_get_gpu_info",
        "tests/unit/recommenders/utils/test_gpu_utils.py::test_get_number_gpus",
        "tests/unit/recommenders/utils/test_gpu_utils.py::test_clear_memory_all_gpus",
        "tests/unit/recommenders/utils/test_gpu_utils.py::test_get_cuda_version",
        "tests/unit/recommenders/utils/test_gpu_utils.py::test_get_cudnn_version",
        "tests/unit/recommenders/utils/test_gpu_utils.py::test_cudnn_enabled",
        "tests/unit/recommenders/utils/test_gpu_utils.py::test_tensorflow_gpu",
        "tests/unit/recommenders/utils/test_gpu_utils.py::test_pytorch_gpu",
        "tests/unit/recommenders/utils/test_tf_utils.py::test_evaluation_log_hook",
        "tests/unit/recommenders/utils/test_tf_utils.py::test_pandas_input_fn",
        "tests/unit/recommenders/utils/test_tf_utils.py::test_pandas_input_fn_for_saved_model",
        "tests/unit/recommenders/utils/test_tf_utils.py::test_build_optimizer",
    ],
    "group_gpu_002": [  # Total group time:
        "tests/unit/examples/test_notebooks_gpu.py::test_gpu_vm",  # 0.76s (Always the first test to check the GPU works)
        # "tests/unit/recommenders/models/test_deeprec_model.py::test_xdeepfm_component_definition",  # FIXME: Disabled due to the issue with TF version > 2.10.1 See #2018
        "tests/unit/recommenders/models/test_deeprec_model.py::test_dkn_component_definition",
        "tests/unit/recommenders/models/test_deeprec_model.py::test_dkn_item2item_component_definition",
        "tests/unit/recommenders/models/test_deeprec_model.py::test_slirec_component_definition",
        "tests/unit/recommenders/models/test_deeprec_model.py::test_nextitnet_component_definition",
        # "tests/unit/recommenders/models/test_deeprec_model.py::test_sum_component_definition",  # FIXME: Disabled due to the issue with TF version > 2.10.1 See #2018
        "tests/unit/recommenders/models/test_deeprec_model.py::test_lightgcn_component_definition",
        "tests/unit/recommenders/models/test_deeprec_utils.py::test_prepare_hparams",
        "tests/unit/recommenders/models/test_deeprec_utils.py::test_load_yaml_file",
        "tests/security/test_dependency_security.py::test_tensorflow",
        "tests/security/test_dependency_security.py::test_torch",
        "tests/regression/test_compatibility_tf.py",
    ],
    "group_notebooks_gpu_001": [  # Total group time: 563.35s
        "tests/unit/examples/test_notebooks_gpu.py::test_gpu_vm",  # 0.76s (Always the first test to check the GPU works)
        "tests/unit/examples/test_notebooks_gpu.py::test_dkn_quickstart",
        "tests/unit/examples/test_notebooks_gpu.py::test_ncf",
        "tests/unit/examples/test_notebooks_gpu.py::test_ncf_deep_dive",
        "tests/unit/examples/test_notebooks_gpu.py::test_fastai",
    ],
    "group_notebooks_gpu_002": [  # Total group time: 241.15s
        "tests/unit/examples/test_notebooks_gpu.py::test_gpu_vm",  # 0.76s (Always the first test to check the GPU works)
        "tests/unit/examples/test_notebooks_gpu.py::test_wide_deep",
        "tests/unit/examples/test_notebooks_gpu.py::test_xdeepfm",
        "tests/unit/examples/test_notebooks_gpu.py::test_gpu_vm",
    ],
}

# Experimental are additional test groups that require to install extra dependencies: pip install .[experimental]
experimental_test_groups = {
    "group_cpu_001": [
        "tests/unit/recommenders/models/test_lightfm_utils.py::test_interactions",
        "tests/unit/recommenders/models/test_lightfm_utils.py::test_fitting",
        "tests/unit/recommenders/models/test_lightfm_utils.py::test_sim_users",
        "tests/unit/recommenders/models/test_lightfm_utils.py::test_sim_items",
        "tests/functional/examples/test_notebooks_python.py::test_lightfm_functional",
        "tests/unit/recommenders/models/test_geoimc.py::test_imcproblem",
        "tests/unit/recommenders/models/test_geoimc.py::test_dataptr",
        "tests/unit/recommenders/models/test_geoimc.py::test_length_normalize",
        "tests/unit/recommenders/models/test_geoimc.py::test_mean_center",
        "tests/unit/recommenders/models/test_geoimc.py::test_reduce_dims",
        "tests/unit/recommenders/models/test_geoimc.py::test_imcproblem",
        "tests/unit/recommenders/models/test_geoimc.py::test_inferer_init",
        "tests/unit/recommenders/models/test_geoimc.py::test_inferer_infer",
        "tests/unit/examples/test_notebooks_python.py::test_rlrmc_quickstart_runs",
    ]
}



# Copyright (c) Recommenders contributors.
# Licensed under the MIT License.

import tensorflow as tf

from recommenders.utils.constants import DEFAULT_USER_COL, DEFAULT_ITEM_COL
from recommenders.utils.tf_utils import MODEL_DIR


def build_feature_columns(
    users,
    items,
    user_col=DEFAULT_USER_COL,
    item_col=DEFAULT_ITEM_COL,
    item_feat_col=None,
    crossed_feat_dim=1000,
    user_dim=8,
    item_dim=8,
    item_feat_shape=None,
    model_type="wide_deep",
):
    """Build wide and/or deep feature columns for TensorFlow high-level API Estimator.

    Args:
        users (iterable): Distinct user ids.
        items (iterable): Distinct item ids.
        user_col (str): User column name.
        item_col (str): Item column name.
        item_feat_col (str): Item feature column name for 'deep' or 'wide_deep' model.
        crossed_feat_dim (int): Crossed feature dimension for 'wide' or 'wide_deep' model.
        user_dim (int): User embedding dimension for 'deep' or 'wide_deep' model.
        item_dim (int): Item embedding dimension for 'deep' or 'wide_deep' model.
        item_feat_shape (int or an iterable of integers): Item feature array shape for 'deep' or 'wide_deep' model.
        model_type (str): Model type, either
            'wide' for a linear model,
            'deep' for a deep neural networks, or
            'wide_deep' for a combination of linear model and neural networks.

    Returns:
        list, list:
        - The wide feature columns
        - The deep feature columns. If only the wide model is selected, the deep column list is empty and viceversa.
    """
    if model_type not in ["wide", "deep", "wide_deep"]:
        raise ValueError("Model type should be either 'wide', 'deep', or 'wide_deep'")

    user_ids = tf.feature_column.categorical_column_with_vocabulary_list(
        user_col, users
    )
    item_ids = tf.feature_column.categorical_column_with_vocabulary_list(
        item_col, items
    )

    if model_type == "wide":
        return _build_wide_columns(user_ids, item_ids, crossed_feat_dim), []
    elif model_type == "deep":
        return (
            [],
            _build_deep_columns(
                user_ids, item_ids, user_dim, item_dim, item_feat_col, item_feat_shape
            ),
        )
    elif model_type == "wide_deep":
        return (
            _build_wide_columns(user_ids, item_ids, crossed_feat_dim),
            _build_deep_columns(
                user_ids, item_ids, user_dim, item_dim, item_feat_col, item_feat_shape
            ),
        )


def _build_wide_columns(user_ids, item_ids, hash_bucket_size=1000):
    """Build wide feature (crossed) columns. `user_ids` * `item_ids` are hashed into `hash_bucket_size`

    Args:
        user_ids (tf.feature_column.categorical_column_with_vocabulary_list): User ids.
        item_ids (tf.feature_column.categorical_column_with_vocabulary_list): Item ids.
        hash_bucket_size (int): Hash bucket size.

    Returns:
        list: Wide feature columns.
    """
    # Including the original features in addition to the crossed one is recommended to address hash collision problem.
    return [
        user_ids,
        item_ids,
        tf.feature_column.crossed_column(
            [user_ids, item_ids], hash_bucket_size=hash_bucket_size
        ),
    ]


def _build_deep_columns(
    user_ids, item_ids, user_dim, item_dim, item_feat_col=None, item_feat_shape=1
):
    """Build deep feature columns

    Args:
        user_ids (tf.feature_column.categorical_column_with_vocabulary_list): User ids.
        item_ids (tf.feature_column.categorical_column_with_vocabulary_list): Item ids.
        user_dim (int): User embedding dimension.
        item_dim (int): Item embedding dimension.
        item_feat_col (str): Item feature column name.
        item_feat_shape (int or an iterable of integers): Item feature array shape.

    Returns:
        list: Deep feature columns.
    """
    deep_columns = [
        # User embedding
        tf.feature_column.embedding_column(
            categorical_column=user_ids, dimension=user_dim, max_norm=user_dim**0.5
        ),
        # Item embedding
        tf.feature_column.embedding_column(
            categorical_column=item_ids, dimension=item_dim, max_norm=item_dim**0.5
        ),
    ]
    # Item feature
    if item_feat_col is not None:
        deep_columns.append(
            tf.feature_column.numeric_column(
                item_feat_col, shape=item_feat_shape, dtype=tf.float32
            )
        )
    return deep_columns


def build_model(
    model_dir=MODEL_DIR,
    wide_columns=(),
    deep_columns=(),
    linear_optimizer="Ftrl",
    dnn_optimizer="Adagrad",
    dnn_hidden_units=(128, 128),
    dnn_dropout=0.0,
    dnn_batch_norm=True,
    log_every_n_iter=1000,
    save_checkpoints_steps=10000,
    seed=None,
):
    """Build wide-deep model.

    To generate wide model, pass wide_columns only.
    To generate deep model, pass deep_columns only.
    To generate wide_deep model, pass both wide_columns and deep_columns.

    Args:
        model_dir (str): Model checkpoint directory.
        wide_columns (list of tf.feature_column): Wide model feature columns.
        deep_columns (list of tf.feature_column): Deep model feature columns.
        linear_optimizer (str or tf.train.Optimizer): Wide model optimizer name or object.
        dnn_optimizer (str or tf.train.Optimizer): Deep model optimizer name or object.
        dnn_hidden_units (list of int): Deep model hidden units. E.g., [10, 10, 10] is three layers of 10 nodes each.
        dnn_dropout (float): Deep model's dropout rate.
        dnn_batch_norm (bool): Deep model's batch normalization flag.
        log_every_n_iter (int): Log the training loss for every n steps.
        save_checkpoints_steps (int): Model checkpoint frequency.
        seed (int): Random seed.

    Returns:
        tf.estimator.Estimator: Model
    """
    gpu_config = tf.compat.v1.ConfigProto()
    gpu_config.gpu_options.allow_growth = True  # dynamic memory allocation

    # TensorFlow training setup
    config = tf.estimator.RunConfig(
        tf_random_seed=seed,
        log_step_count_steps=log_every_n_iter,
        save_checkpoints_steps=save_checkpoints_steps,
        session_config=gpu_config,
    )

    if len(wide_columns) > 0 and len(deep_columns) == 0:
        model = tf.compat.v1.estimator.LinearRegressor(
            model_dir=model_dir,
            config=config,
            feature_columns=wide_columns,
            optimizer=linear_optimizer,
        )
    elif len(wide_columns) == 0 and len(deep_columns) > 0:
        model = tf.compat.v1.estimator.DNNRegressor(
            model_dir=model_dir,
            config=config,
            feature_columns=deep_columns,
            hidden_units=dnn_hidden_units,
            optimizer=dnn_optimizer,
            dropout=dnn_dropout,
            batch_norm=dnn_batch_norm,
        )
    elif len(wide_columns) > 0 and len(deep_columns) > 0:
        model = tf.compat.v1.estimator.DNNLinearCombinedRegressor(
            model_dir=model_dir,
            config=config,
            # wide settings
            linear_feature_columns=wide_columns,
            linear_optimizer=linear_optimizer,
            # deep settings
            dnn_feature_columns=deep_columns,
            dnn_hidden_units=dnn_hidden_units,
            dnn_optimizer=dnn_optimizer,
            dnn_dropout=dnn_dropout,
            batch_norm=dnn_batch_norm,
        )
    else:
        raise ValueError(
            "To generate wide model, set wide_columns.\n"
            "To generate deep model, set deep_columns.\n"
            "To generate wide_deep model, set both wide_columns and deep_columns."
        )

    return model



# Copyright (c) Recommenders contributors.
# Licensed under the MIT License.
# Original codes are from
# https://github.com/kang205/SASRec/blob/master/sampler.py

import numpy as np
from multiprocessing import Process, Queue


def random_neq(left, right, s):
    t = np.random.randint(left, right)
    while t in s:
        t = np.random.randint(left, right)
    return t


def sample_function(
    user_train, usernum, itemnum, batch_size, maxlen, result_queue, seed
):
    """Batch sampler that creates a sequence of negative items based on the
    original sequence of items (positive) that the user has interacted with.

    Args:
        user_train (dict): dictionary of training exampled for each user
        usernum (int): number of users
        itemnum (int): number of items
        batch_size (int): batch size
        maxlen (int): maximum input sequence length
        result_queue (multiprocessing.Queue): queue for storing sample results
        seed (int): seed for random generator
    """

    def sample():

        user = np.random.randint(1, usernum + 1)
        while len(user_train[user]) <= 1:
            user = np.random.randint(1, usernum + 1)

        seq = np.zeros([maxlen], dtype=np.int32)
        pos = np.zeros([maxlen], dtype=np.int32)
        neg = np.zeros([maxlen], dtype=np.int32)
        nxt = user_train[user][-1]
        idx = maxlen - 1

        ts = set(user_train[user])
        for i in reversed(user_train[user][:-1]):
            seq[idx] = i
            pos[idx] = nxt
            if nxt != 0:
                neg[idx] = random_neq(1, itemnum + 1, ts)
            nxt = i
            idx -= 1
            if idx == -1:
                break

        return (user, seq, pos, neg)

    np.random.seed(seed)
    while True:
        one_batch = []
        for i in range(batch_size):
            one_batch.append(sample())

        result_queue.put(zip(*one_batch))


class WarpSampler(object):
    """Sampler object that creates an iterator for feeding batch data while training.

    Attributes:
        User: dict, all the users (keys) with items as values
        usernum: integer, total number of users
        itemnum: integer, total number of items
        batch_size (int): batch size
        maxlen (int): maximum input sequence length
        n_workers (int): number of workers for parallel execution
    """

    def __init__(self, User, usernum, itemnum, batch_size=64, maxlen=10, n_workers=1):
        self.result_queue = Queue(maxsize=n_workers * 10)
        self.processors = []
        for i in range(n_workers):
            self.processors.append(
                Process(
                    target=sample_function,
                    args=(
                        User,
                        usernum,
                        itemnum,
                        batch_size,
                        maxlen,
                        self.result_queue,
                        np.random.randint(2e9),
                    ),
                )
            )
            self.processors[-1].daemon = True
            self.processors[-1].start()

    def next_batch(self):
        return self.result_queue.get()

    def close(self):
        for p in self.processors:
            p.terminate()
            p.join()



# Copyright (c) Recommenders contributors.
# Licensed under the MIT License.


import sys
import pytest

try:
    import tensorflow as tf
    import torch
    from recommenders.utils.gpu_utils import (
        get_cuda_version,
        get_cudnn_version,
        get_gpu_info,
        get_number_gpus,
    )
except ImportError:
    pass  # skip this import if we are in cpu environment


@pytest.mark.gpu
def test_get_gpu_info():
    assert len(get_gpu_info()) >= 1


@pytest.mark.gpu
def test_get_number_gpus():
    assert get_number_gpus() >= 1


@pytest.mark.gpu
@pytest.mark.skip(reason="TODO: Implement this")
def test_clear_memory_all_gpus():
    pass


@pytest.mark.gpu
@pytest.mark.skipif(sys.platform == "win32", reason="Not implemented on Windows")
def test_get_cuda_version():
    assert int(get_cuda_version().split(".")[0]) > 9


@pytest.mark.gpu
def test_get_cudnn_version():
    assert int(get_cudnn_version()[0]) > 7


@pytest.mark.gpu
def test_cudnn_enabled():
    assert torch.backends.cudnn.enabled is True


@pytest.mark.gpu
@pytest.mark.skip(reason="This function in TF is flaky")
def test_tensorflow_gpu():
    assert len(tf.config.list_physical_devices("GPU")) > 0


@pytest.mark.gpu
@pytest.mark.skip(reason="This function in PyTorch is flaky")
def test_pytorch_gpu():
    assert torch.cuda.is_available()



# Copyright (c) Recommenders contributors.
# Licensed under the MIT License.

import tensorflow as tf
from recommenders.models.deeprec.models.sequential.sequential_base_model import (
    SequentialBaseModel,
)

__all__ = ["NextItNetModel"]


class NextItNetModel(SequentialBaseModel):
    """NextItNet Model

    :Citation:
        Yuan, Fajie, et al. "A Simple Convolutional Generative Network
        for Next Item Recommendation", in Web Search and Data Mining, 2019.

    Note:
        It requires strong sequence with dataset.
    """

    def _build_seq_graph(self):
        """The main function to create nextitnet model.

        Returns:
            object: The output of nextitnet section.
        """
        hparams = self.hparams
        is_training = tf.equal(self.is_train_stage, True)
        item_history_embedding = tf.cond(
            pred=is_training,
            true_fn=lambda: self.item_history_embedding[
                :: self.hparams.train_num_ngs + 1
            ],
            false_fn=lambda: self.item_history_embedding,
        )
        cate_history_embedding = tf.cond(
            pred=is_training,
            true_fn=lambda: self.cate_history_embedding[
                :: self.hparams.train_num_ngs + 1
            ],
            false_fn=lambda: self.cate_history_embedding,
        )

        with tf.compat.v1.variable_scope("nextitnet", reuse=tf.compat.v1.AUTO_REUSE):
            dilate_input = tf.concat(
                [item_history_embedding, cate_history_embedding], 2
            )

            for layer_id, dilation in enumerate(hparams.dilations):
                dilate_input = tf.cond(
                    pred=is_training,
                    true_fn=lambda: self._nextitnet_residual_block_one(
                        dilate_input,
                        dilation,
                        layer_id,
                        dilate_input.get_shape()[-1],
                        hparams.kernel_size,
                        causal=True,
                        train=True,
                    ),
                    false_fn=lambda: self._nextitnet_residual_block_one(
                        dilate_input,
                        dilation,
                        layer_id,
                        dilate_input.get_shape()[-1],
                        hparams.kernel_size,
                        causal=True,
                        train=False,
                    ),
                )

            self.dilate_input = dilate_input
            model_output = tf.cond(
                pred=is_training,
                true_fn=self._training_output,
                false_fn=self._normal_output,
            )

            return model_output

    def _training_output(self):
        model_output = tf.repeat(
            self.dilate_input, self.hparams.train_num_ngs + 1, axis=0
        )
        model_output = tf.concat([model_output, self.target_item_embedding], -1)
        model_output = tf.reshape(
            model_output,
            (
                -1,
                self.hparams.train_num_ngs + 1,
                self.hparams.max_seq_length,
                model_output.get_shape()[-1],
            ),
        )
        model_output = tf.transpose(a=model_output, perm=[0, 2, 1, 3])
        model_output = tf.reshape(model_output, (-1, model_output.get_shape()[-1]))
        return model_output

    def _normal_output(self):
        model_output = self.dilate_input[:, -1, :]
        model_output = tf.concat(
            [model_output, self.target_item_embedding[:, -1, :]], -1
        )
        return model_output

    def _nextitnet_residual_block_one(
        self,
        input_,
        dilation,
        layer_id,
        residual_channels,
        kernel_size,
        causal=True,
        train=True,
    ):
        """The main function to use dilated CNN and residual network at sequence data

        Args:
            input_ (object): The output of history sequential embeddings
            dilation (int): The dilation number of CNN layer
            layer_id (str): String value of layer ID, 0, 1, 2...
            residual_channels (int): Embedding size of input sequence
            kernel_size (int): Kernel size of CNN mask
            causal (bool): Whether to pad in front of the sequence or to pad surroundingly
            train (bool): is in training stage

        Returns:
            object: The output of residual layers.
        """
        resblock_type = "decoder"
        resblock_name = "nextitnet_residual_block_one_{}_layer_{}_{}".format(
            resblock_type, layer_id, dilation
        )
        with tf.compat.v1.variable_scope(resblock_name):
            input_ln = self._layer_norm(input_, name="layer_norm1", trainable=train)
            relu1 = tf.nn.relu(input_ln)
            conv1 = self._conv1d(
                relu1, int(0.5 * int(residual_channels)), name="conv1d_1"
            )
            conv1 = self._layer_norm(conv1, name="layer_norm2", trainable=train)
            relu2 = tf.nn.relu(conv1)

            dilated_conv = self._conv1d(
                relu2,
                int(0.5 * int(residual_channels)),
                dilation,
                kernel_size,
                causal=causal,
                name="dilated_conv",
            )

            dilated_conv = self._layer_norm(
                dilated_conv, name="layer_norm3", trainable=train
            )
            relu3 = tf.nn.relu(dilated_conv)
            conv2 = self._conv1d(relu3, residual_channels, name="conv1d_2")
            return input_ + conv2

    def _conv1d(
        self,
        input_,
        output_channels,
        dilation=1,
        kernel_size=1,
        causal=False,
        name="dilated_conv",
    ):
        """Call a dilated CNN layer

        Returns:
            object: The output of dilated CNN layers.
        """
        with tf.compat.v1.variable_scope(name):
            weight = tf.compat.v1.get_variable(
                "weight",
                [1, kernel_size, input_.get_shape()[-1], output_channels],
                initializer=tf.compat.v1.truncated_normal_initializer(
                    stddev=0.02, seed=1
                ),
            )
            bias = tf.compat.v1.get_variable(
                "bias",
                [output_channels],
                initializer=tf.compat.v1.constant_initializer(0.0),
            )

            if causal:
                padding = [[0, 0], [(kernel_size - 1) * dilation, 0], [0, 0]]
                padded = tf.pad(tensor=input_, paddings=padding)
                input_expanded = tf.expand_dims(padded, axis=1)
                out = (
                    tf.nn.atrous_conv2d(
                        input_expanded, weight, rate=dilation, padding="VALID"
                    )
                    + bias
                )
            else:
                input_expanded = tf.expand_dims(input_, axis=1)
                out = (
                    tf.nn.conv2d(
                        input=input_expanded,
                        filters=weight,
                        strides=[1, 1, 1, 1],
                        padding="SAME",
                    )
                    + bias
                )

            return tf.squeeze(out, [1])

    def _layer_norm(self, x, name, epsilon=1e-8, trainable=True):
        """Call a layer normalization

        Returns:
            object: Normalized data
        """
        with tf.compat.v1.variable_scope(name):
            shape = x.get_shape()
            beta = tf.compat.v1.get_variable(
                "beta",
                [int(shape[-1])],
                initializer=tf.compat.v1.constant_initializer(0),
                trainable=trainable,
            )
            gamma = tf.compat.v1.get_variable(
                "gamma",
                [int(shape[-1])],
                initializer=tf.compat.v1.constant_initializer(1),
                trainable=trainable,
            )

            mean, variance = tf.nn.moments(x=x, axes=[len(shape) - 1], keepdims=True)

            x = (x - mean) / tf.sqrt(variance + epsilon)

            return gamma * x + beta



# Copyright (c) Recommenders contributors.
# Licensed under the MIT License.

import tensorflow as tf
from recommenders.models.deeprec.models.sequential.sequential_base_model import (
    SequentialBaseModel,
)

__all__ = ["CaserModel"]


class CaserModel(SequentialBaseModel):
    """Caser Model

    :Citation:

        J. Tang and K. Wang, "Personalized top-n sequential recommendation via convolutional
        sequence embedding", in Proceedings of the Eleventh ACM International Conference on
        Web Search and Data Mining, ACM, 2018.
    """

    def __init__(self, hparams, iterator_creator, seed=None):
        """Initialization of variables for caser

        Args:
            hparams (HParams): A HParams object, hold the entire set of hyperparameters.
            iterator_creator (object): An iterator to load the data.
        """
        self.hparams = hparams
        self.L = hparams.L  # history sequence that involved in convolution shape
        self.T = hparams.T  # prediction shape
        self.n_v = hparams.n_v  # number of vertical convolution layers
        self.n_h = hparams.n_h  # number of horizonal convolution layers
        self.lengths = [
            i + 1 for i in range(self.L)
        ]  # horizonal convolution filter shape
        super().__init__(hparams, iterator_creator, seed=seed)

    def _build_seq_graph(self):
        """The main function to create caser model.

        Returns:
            object: The output of caser section.
        """
        with tf.compat.v1.variable_scope("caser"):
            cnn_output = self._caser_cnn()
            model_output = tf.concat([cnn_output, self.target_item_embedding], 1)
            tf.compat.v1.summary.histogram("model_output", model_output)
            return model_output

    def _add_cnn(self, hist_matrix, vertical_dim, scope):
        """The main function to use CNN at both vertical and horizonal aspects.

        Args:
            hist_matrix (object): The output of history sequential embeddings
            vertical_dim (int): The shape of embeddings of input
            scope (object): The scope of CNN input.

        Returns:
            object: The output of CNN layers.
        """
        with tf.compat.v1.variable_scope(scope):
            with tf.compat.v1.variable_scope("vertical"):
                embedding_T = tf.transpose(a=hist_matrix, perm=[0, 2, 1])
                out_v = self._build_cnn(embedding_T, self.n_v, vertical_dim)
                out_v = tf.compat.v1.layers.flatten(out_v)
            with tf.compat.v1.variable_scope("horizonal"):
                out_hs = []
                for h in self.lengths:
                    conv_out = self._build_cnn(hist_matrix, self.n_h, h)
                    max_pool_out = tf.reduce_max(
                        input_tensor=conv_out, axis=[1], name="max_pool_{0}".format(h)
                    )
                    out_hs.append(max_pool_out)
                out_h = tf.concat(out_hs, 1)
        return tf.concat([out_v, out_h], 1)

    def _caser_cnn(self):
        """The main function to use CNN at both item and category aspects.

        Returns:
            object: The concatenated output of two parts of item and category.
        """
        item_out = self._add_cnn(
            self.item_history_embedding, self.item_embedding_dim, "item"
        )
        tf.compat.v1.summary.histogram("item_out", item_out)
        cate_out = self._add_cnn(
            self.cate_history_embedding, self.cate_embedding_dim, "cate"
        )
        tf.compat.v1.summary.histogram("cate_out", cate_out)
        cnn_output = tf.concat([item_out, cate_out], 1)
        tf.compat.v1.summary.histogram("cnn_output", cnn_output)
        return cnn_output

    def _build_cnn(self, history_matrix, nums, shape):
        """Call a CNN layer.

        Returns:
            object: The output of cnn section.
        """
        return tf.compat.v1.layers.conv1d(
            history_matrix,
            nums,
            shape,
            activation=tf.nn.relu,
            name="conv_" + str(shape),
        )



# Copyright (c) Recommenders contributors.
# Licensed under the MIT License.


import os
import pytest

try:
    from recommenders.models.newsrec.newsrec_utils import prepare_hparams
    from recommenders.models.deeprec.deeprec_utils import download_deeprec_resources
    from recommenders.models.newsrec.io.mind_iterator import MINDIterator
    from recommenders.models.newsrec.io.mind_all_iterator import MINDAllIterator
except ImportError:
    pass  # disable error while collecting tests for non-gpu environments


@pytest.mark.gpu
def test_news_iterator(mind_resource_path):
    train_news_file = os.path.join(mind_resource_path, "train", "news.tsv")
    train_behaviors_file = os.path.join(mind_resource_path, "train", "behaviors.tsv")
    valid_news_file = os.path.join(mind_resource_path, "valid", "news.tsv")
    valid_behaviors_file = os.path.join(mind_resource_path, "valid", "behaviors.tsv")
    wordEmb_file = os.path.join(mind_resource_path, "utils", "embedding.npy")
    userDict_file = os.path.join(mind_resource_path, "utils", "uid2index.pkl")
    wordDict_file = os.path.join(mind_resource_path, "utils", "word_dict.pkl")
    yaml_file = os.path.join(mind_resource_path, "utils", "nrms.yaml")

    if not os.path.exists(train_news_file):
        download_deeprec_resources(
            "https://recodatasets.z20.web.core.windows.net/newsrec/",
            os.path.join(mind_resource_path, "train"),
            "MINDdemo_train.zip",
        )
    if not os.path.exists(valid_news_file):
        download_deeprec_resources(
            "https://recodatasets.z20.web.core.windows.net/newsrec/",
            os.path.join(mind_resource_path, "valid"),
            "MINDdemo_dev.zip",
        )
    if not os.path.exists(yaml_file):
        download_deeprec_resources(
            "https://recodatasets.z20.web.core.windows.net/newsrec/",
            os.path.join(mind_resource_path, "utils"),
            "MINDdemo_utils.zip",
        )

    hparams = prepare_hparams(
        yaml_file,
        wordEmb_file=wordEmb_file,
        wordDict_file=wordDict_file,
        userDict_file=userDict_file,
        epochs=1,
    )
    train_iterator = MINDIterator(hparams, hparams.npratio)
    test_iterator = MINDIterator(hparams, -1)

    assert train_iterator is not None
    for res in train_iterator.load_data_from_file(
        train_news_file, train_behaviors_file
    ):
        assert isinstance(res, dict)
        assert len(res) == 5
        break

    assert test_iterator is not None
    for res in test_iterator.load_data_from_file(valid_news_file, valid_behaviors_file):
        assert isinstance(res, dict)
        assert len(res) == 5
        break


@pytest.mark.gpu
def test_naml_iterator(mind_resource_path):
    train_news_file = os.path.join(mind_resource_path, "train", "news.tsv")
    train_behaviors_file = os.path.join(mind_resource_path, "train", "behaviors.tsv")
    valid_news_file = os.path.join(mind_resource_path, "valid", "news.tsv")
    valid_behaviors_file = os.path.join(mind_resource_path, "valid", "behaviors.tsv")
    wordEmb_file = os.path.join(mind_resource_path, "utils", "embedding_all.npy")
    userDict_file = os.path.join(mind_resource_path, "utils", "uid2index.pkl")
    wordDict_file = os.path.join(mind_resource_path, "utils", "word_dict_all.pkl")
    vertDict_file = os.path.join(mind_resource_path, "utils", "vert_dict.pkl")
    subvertDict_file = os.path.join(mind_resource_path, "utils", "subvert_dict.pkl")
    yaml_file = os.path.join(mind_resource_path, "utils", "naml.yaml")

    if not os.path.exists(train_news_file):
        download_deeprec_resources(
            "https://recodatasets.z20.web.core.windows.net/newsrec/",
            os.path.join(mind_resource_path, "train"),
            "MINDdemo_train.zip",
        )
    if not os.path.exists(valid_news_file):
        download_deeprec_resources(
            "https://recodatasets.z20.web.core.windows.net/newsrec/",
            os.path.join(mind_resource_path, "valid"),
            "MINDdemo_dev.zip",
        )
    if not os.path.exists(yaml_file):
        download_deeprec_resources(
            "https://recodatasets.z20.web.core.windows.net/newsrec/",
            os.path.join(mind_resource_path, "utils"),
            "MINDdemo_utils.zip",
        )

    hparams = prepare_hparams(
        yaml_file,
        wordEmb_file=wordEmb_file,
        wordDict_file=wordDict_file,
        userDict_file=userDict_file,
        vertDict_file=vertDict_file,
        subvertDict_file=subvertDict_file,
        epochs=1,
        batch_size=1024,
    )
    train_iterator = MINDAllIterator(hparams, hparams.npratio)
    test_iterator = MINDAllIterator(hparams, -1)

    assert train_iterator is not None
    for res in train_iterator.load_data_from_file(
        train_news_file, train_behaviors_file
    ):
        assert isinstance(res, dict)
        assert len(res) == 11
        break

    assert test_iterator is not None
    for res in test_iterator.load_data_from_file(valid_news_file, valid_behaviors_file):
        assert isinstance(res, dict)
        assert len(res) == 11
        break



# Copyright (c) Recommenders contributors.
# Licensed under the MIT License.

import os
import logging
import requests
import math
import zipfile
from contextlib import contextmanager
from tempfile import TemporaryDirectory
from tqdm import tqdm
from retrying import retry


log = logging.getLogger(__name__)


@retry(wait_random_min=1000, wait_random_max=5000, stop_max_attempt_number=5)
def maybe_download(url, filename=None, work_directory=".", expected_bytes=None):
    """Download a file if it is not already downloaded.

    Args:
        filename (str): File name.
        work_directory (str): Working directory.
        url (str): URL of the file to download.
        expected_bytes (int): Expected file size in bytes.

    Returns:
        str: File path of the file downloaded.
    """
    if filename is None:
        filename = url.split("/")[-1]
    os.makedirs(work_directory, exist_ok=True)
    filepath = os.path.join(work_directory, filename)
    if not os.path.exists(filepath):
        r = requests.get(url, stream=True)
        if r.status_code == 200:
            log.info(f"Downloading {url}")
            total_size = int(r.headers.get("content-length", 0))
            block_size = 1024
            num_iterables = math.ceil(total_size / block_size)
            with open(filepath, "wb") as file:
                for data in tqdm(
                    r.iter_content(block_size),
                    total=num_iterables,
                    unit="KB",
                    unit_scale=True,
                ):
                    file.write(data)
        else:
            log.error(f"Problem downloading {url}")
            r.raise_for_status()
    else:
        log.info(f"File {filepath} already downloaded")
    if expected_bytes is not None:
        statinfo = os.stat(filepath)
        if statinfo.st_size != expected_bytes:
            os.remove(filepath)
            raise IOError(f"Failed to verify {filepath}")

    return filepath


@contextmanager
def download_path(path=None):
    """Return a path to download data. If `path=None`, then it yields a temporal path that is eventually deleted,
    otherwise the real path of the input.

    Args:
        path (str): Path to download data.

    Returns:
        str: Real path where the data is stored.

    Examples:
        >>> with download_path() as path:
        >>> ... maybe_download(url="http://example.com/file.zip", work_directory=path)

    """
    if path is None:
        tmp_dir = TemporaryDirectory()
        try:
            yield tmp_dir.name
        finally:
            tmp_dir.cleanup()
    else:
        path = os.path.realpath(path)
        yield path


def unzip_file(zip_src, dst_dir, clean_zip_file=False):
    """Unzip a file

    Args:
        zip_src (str): Zip file.
        dst_dir (str): Destination folder.
        clean_zip_file (bool): Whether or not to clean the zip file.
    """
    fz = zipfile.ZipFile(zip_src, "r")
    for file in fz.namelist():
        fz.extract(file, dst_dir)
    if clean_zip_file:
        os.remove(zip_src)



# Copyright (c) Recommenders contributors.
# Licensed under the MIT License.

import tensorflow as tf
import numpy as np
import random

from recommenders.models.deeprec.io.sequential_iterator import SequentialIterator
from recommenders.models.deeprec.deeprec_utils import load_dict


__all__ = ["NextItNetIterator"]


class NextItNetIterator(SequentialIterator):
    """Data loader for the NextItNet model.

    NextItNet requires a special type of data format. In training stage, each instance will
    produce `(sequence_length * train_num_ngs)` target items and labels, to let NextItNet
    output predictions of every item in a sequence except only of the last item.
    """

    def __init__(self, hparams, graph, col_spliter="\t"):
        """Initialize an iterator. Create necessary placeholders for the model.
        Different from sequential iterator

        Args:
            hparams (object): Global hyper-parameters. Some key settings such as #_feature and #_field are there.
            graph (object): The running graph. All created placeholder will be added to this graph.
            col_spliter (str): Column splitter in one line.
        """
        self.col_spliter = col_spliter

        self.userdict, self.itemdict, self.catedict = (
            load_dict(hparams.user_vocab),
            load_dict(hparams.item_vocab),
            load_dict(hparams.cate_vocab),
        )

        self.max_seq_length = hparams.max_seq_length
        self.batch_size = hparams.batch_size
        self.iter_data = dict()

        self.graph = graph
        with self.graph.as_default():
            self.labels = tf.compat.v1.placeholder(
                tf.float32, [None, None], name="label"
            )
            self.users = tf.compat.v1.placeholder(tf.int32, [None], name="users")
            self.items = tf.compat.v1.placeholder(tf.int32, [None, None], name="items")
            self.cates = tf.compat.v1.placeholder(tf.int32, [None, None], name="cates")
            self.item_history = tf.compat.v1.placeholder(
                tf.int32, [None, self.max_seq_length], name="item_history"
            )
            self.item_cate_history = tf.compat.v1.placeholder(
                tf.int32, [None, self.max_seq_length], name="item_cate_history"
            )
            self.mask = tf.compat.v1.placeholder(
                tf.int32, [None, self.max_seq_length], name="mask"
            )
            self.time = tf.compat.v1.placeholder(tf.float32, [None], name="time")
            self.time_diff = tf.compat.v1.placeholder(
                tf.float32, [None, self.max_seq_length], name="time_diff"
            )
            self.time_from_first_action = tf.compat.v1.placeholder(
                tf.float32, [None, self.max_seq_length], name="time_from_first_action"
            )
            self.time_to_now = tf.compat.v1.placeholder(
                tf.float32, [None, self.max_seq_length], name="time_to_now"
            )

    def _convert_data(
        self,
        label_list,
        user_list,
        item_list,
        item_cate_list,
        item_history_batch,
        item_cate_history_batch,
        time_list,
        time_diff_list,
        time_from_first_action_list,
        time_to_now_list,
        batch_num_ngs,
    ):
        """Convert data into numpy arrays that are good for further model operation.
        Note: This is different from `sequential_iterator`.

        Args:
            label_list (list): A list of ground-truth labels.
            user_list (list): A list of user indexes.
            item_list (list): A list of item indexes.
            item_cate_list (list): A list of category indexes.
            item_history_batch (list): A list of item history indexes.
            item_cate_history_batch (list): A list of category history indexes.
            time_list (list): A list of current timestamp.
            time_diff_list (list): A list of timestamp between each sequential opertions.
            time_from_first_action_list (list): A list of timestamp from the first opertion.
            time_to_now_list (list): A list of timestamp to the current time.
            batch_num_ngs (int): The number of negative sampling while training in mini-batch.

        Returns:
            dict: A dictionary, contains multiple numpy arrays that are convenient for further operation.
        """
        if batch_num_ngs:
            instance_cnt = len(label_list)
            if instance_cnt < 5:
                return

            label_list_all = []
            item_list_all = []
            item_cate_list_all = []
            user_list_all = np.asarray(
                [[user] * (batch_num_ngs + 1) for user in user_list], dtype=np.int32
            ).flatten()
            time_list_all = np.asarray(
                [[t] * (batch_num_ngs + 1) for t in time_list], dtype=np.float32
            ).flatten()

            history_lengths = [len(item_history_batch[i]) for i in range(instance_cnt)]
            max_seq_length_batch = self.max_seq_length
            item_history_batch_all = np.zeros(
                (instance_cnt * (batch_num_ngs + 1), max_seq_length_batch),
                dtype=np.int32,
            )
            item_cate_history_batch_all = np.zeros(
                (instance_cnt * (batch_num_ngs + 1), max_seq_length_batch),
                dtype=np.int32,
            )
            time_diff_batch = np.zeros(
                (instance_cnt * (batch_num_ngs + 1), max_seq_length_batch),
                dtype=np.float32,
            )
            time_from_first_action_batch = np.zeros(
                (instance_cnt * (batch_num_ngs + 1), max_seq_length_batch),
                dtype=np.float32,
            )
            time_to_now_batch = np.zeros(
                (instance_cnt * (batch_num_ngs + 1), max_seq_length_batch),
                dtype=np.float32,
            )
            mask = np.zeros(
                (instance_cnt * (1 + batch_num_ngs), max_seq_length_batch),
                dtype=np.float32,
            )

            for i in range(instance_cnt):
                this_length = min(history_lengths[i], max_seq_length_batch)
                for index in range(batch_num_ngs + 1):
                    item_history_batch_all[
                        i * (batch_num_ngs + 1) + index, -this_length:
                    ] = np.asarray(item_history_batch[i][-this_length:], dtype=np.int32)
                    item_cate_history_batch_all[
                        i * (batch_num_ngs + 1) + index, -this_length:
                    ] = np.asarray(
                        item_cate_history_batch[i][-this_length:], dtype=np.int32
                    )
                    mask[i * (batch_num_ngs + 1) + index, -this_length:] = 1.0
                    time_diff_batch[
                        i * (batch_num_ngs + 1) + index, -this_length:
                    ] = np.asarray(time_diff_list[i][-this_length:], dtype=np.float32)
                    time_from_first_action_batch[
                        i * (batch_num_ngs + 1) + index, -this_length:
                    ] = np.asarray(
                        time_from_first_action_list[i][-this_length:], dtype=np.float32
                    )
                    time_to_now_batch[
                        i * (batch_num_ngs + 1) + index, -this_length:
                    ] = np.asarray(time_to_now_list[i][-this_length:], dtype=np.float32)

            for i in range(instance_cnt):
                positive_item = [
                    *item_history_batch_all[i * (batch_num_ngs + 1)][1:],
                    item_list[i],
                ]
                positive_item_cate = [
                    *item_cate_history_batch_all[i * (batch_num_ngs + 1)][1:],
                    item_cate_list[i],
                ]
                label_list_all.append([1] * max_seq_length_batch)
                item_list_all.append(positive_item)
                item_cate_list_all.append(positive_item_cate)

                count = 0
                while count < batch_num_ngs:
                    negative_item_list = []
                    negative_item_cate_list = []
                    count_inner = 1
                    while count_inner <= max_seq_length_batch:
                        random_value = random.randint(0, instance_cnt - 1)
                        negative_item = item_list[random_value]
                        if negative_item == positive_item[count_inner - 1]:
                            continue
                        negative_item_list.append(negative_item)
                        negative_item_cate_list.append(item_cate_list[random_value])
                        count_inner += 1

                    label_list_all.append([0] * max_seq_length_batch)
                    item_list_all.append(negative_item_list)
                    item_cate_list_all.append(negative_item_cate_list)
                    count += 1

            res = {}
            res["labels"] = np.asarray(
                label_list_all, dtype=np.float32
            )  # .reshape(-1,1)
            res["users"] = user_list_all
            res["items"] = np.asarray(item_list_all, dtype=np.int32)
            res["cates"] = np.asarray(item_cate_list_all, dtype=np.int32)
            res["item_history"] = item_history_batch_all
            res["item_cate_history"] = item_cate_history_batch_all
            res["mask"] = mask
            res["time"] = time_list_all
            res["time_diff"] = time_diff_batch
            res["time_from_first_action"] = time_from_first_action_batch
            res["time_to_now"] = time_to_now_batch

            return res

        else:
            instance_cnt = len(label_list)
            history_lengths = [len(item_history_batch[i]) for i in range(instance_cnt)]
            max_seq_length_batch = self.max_seq_length
            item_history_batch_all = np.zeros(
                (instance_cnt, max_seq_length_batch), dtype=np.int32
            )
            item_cate_history_batch_all = np.zeros(
                (instance_cnt, max_seq_length_batch), dtype=np.int32
            )
            time_diff_batch = np.zeros(
                (instance_cnt, max_seq_length_batch), dtype=np.float32
            )
            time_from_first_action_batch = np.zeros(
                (instance_cnt, max_seq_length_batch), dtype=np.float32
            )
            time_to_now_batch = np.zeros(
                (instance_cnt, max_seq_length_batch), dtype=np.float32
            )
            mask = np.zeros((instance_cnt, max_seq_length_batch), dtype=np.float32)

            for i in range(instance_cnt):
                this_length = min(history_lengths[i], max_seq_length_batch)
                item_history_batch_all[i, -this_length:] = item_history_batch[i][
                    -this_length:
                ]
                item_cate_history_batch_all[i, -this_length:] = item_cate_history_batch[
                    i
                ][-this_length:]
                mask[i, -this_length:] = 1.0
                time_diff_batch[i, -this_length:] = time_diff_list[i][-this_length:]
                time_from_first_action_batch[
                    i, -this_length:
                ] = time_from_first_action_list[i][-this_length:]
                time_to_now_batch[i, -this_length:] = time_to_now_list[i][-this_length:]

            res = {}
            res["labels"] = np.asarray(label_list, dtype=np.float32).reshape([-1, 1])
            res["users"] = np.asarray(user_list, dtype=np.float32)
            res["items"] = np.asarray(item_list, dtype=np.int32).reshape([-1, 1])
            res["cates"] = np.asarray(item_cate_list, dtype=np.int32).reshape([-1, 1])
            res["item_history"] = item_history_batch_all
            res["item_cate_history"] = item_cate_history_batch_all
            res["mask"] = mask
            res["time"] = np.asarray(time_list, dtype=np.float32)
            res["time_diff"] = time_diff_batch
            res["time_from_first_action"] = time_from_first_action_batch
            res["time_to_now"] = time_to_now_batch
            return res
